# flake8: noqa
# Copyright 2020 IBM
# Author: peter.zhong@au1.ibm.com
#
# This is free software; you can redistribute it and/or modify
# it under the terms of the Apache 2.0 License.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache 2.0 License for more details.

import ast
import distance
import editdistance
import json
import Levenshtein
import numpy as np
import re
import string
from apted import APTED, Config
from apted.helpers import Tree
from collections import deque
from itertools import product
from lxml import etree, html
from tqdm import tqdm
from typing import Any, Callable, Optional, Sequence
from zss import Node, simple_distance

from .parallel import parallel_process


class TableTree(Tree):

    def __init__(self, tag, colspan=None, rowspan=None, content=None, *children):
        self.tag = tag
        self.colspan = colspan
        self.rowspan = rowspan
        self.content = content
        self.children = list(children)

    def bracket(self):
        """Show tree using brackets notation"""
        if self.tag == 'td':
            result = '"tag": %s, "colspan": %d, "rowspan": %d, "text": %s' % (
                self.tag, self.colspan, self.rowspan, self.content
            )
        else:
            result = '"tag": %s' % self.tag
        for child in self.children:
            result += child.bracket()
        return '{{{}}}'.format(result)


class CustomConfig(Config):

    @staticmethod
    def maximum(*sequences):
        """Get maximum possible value"""
        return max(map(len, sequences))

    def normalized_distance(self, *sequences):
        """Get distance from 0 to 1"""
        return float(distance.levenshtein(*sequences)) / self.maximum(*sequences)

    def rename(self, node1, node2):
        """Compares attributes of trees"""
        if (node1.tag != node2.tag) or (node1.colspan != node2.colspan) or (node1.rowspan != node2.rowspan):
            return 1.0
        if node1.tag == 'td':
            if node1.content or node2.content:
                return self.normalized_distance(node1.content, node2.content)
        return 0.0


class TEDS(object):
    """Tree Edit Distance basead Similarity"""

    def __init__(self, structure_only=False, n_jobs=1, ignore_nodes=None):
        assert isinstance(n_jobs, int) and (n_jobs >= 1), 'n_jobs must be an integer greather than 1'
        self.structure_only = structure_only
        self.n_jobs = n_jobs
        self.ignore_nodes = ignore_nodes
        self.__tokens__ = []

    def tokenize(self, node):
        """Tokenizes table cells"""
        self.__tokens__.append('<%s>' % node.tag)
        if node.text is not None:
            self.__tokens__ += list(node.text)
        for n in node.getchildren():
            self.tokenize(n)
        if node.tag != 'unk':
            self.__tokens__.append('</%s>' % node.tag)
        if node.tag != 'td' and node.tail is not None:
            self.__tokens__ += list(node.tail)

    def load_html_tree(self, node, parent=None):
        """Converts HTML tree to the format required by apted"""
        global __tokens__
        if node.tag == 'td':
            if self.structure_only:
                cell = []
            else:
                self.__tokens__ = []
                self.tokenize(node)
                cell = self.__tokens__[1:-1].copy()
            new_node = TableTree(
                node.tag, int(node.attrib.get('colspan', '1')), int(node.attrib.get('rowspan', '1')), cell, *deque()
            )
        else:
            new_node = TableTree(node.tag, None, None, None, *deque())
        if parent is not None:
            parent.children.append(new_node)
        if node.tag != 'td':
            for n in node.getchildren():
                self.load_html_tree(n, new_node)
        if parent is None:
            return new_node

    def evaluate(self, pred, true):
        """Computes TEDS score between the prediction and the ground truth of a
        given sample
        """
        if (not pred) or (not true):
            return 0.0
        parser = html.HTMLParser(remove_comments=True, encoding='utf-8')
        pred = html.fromstring(pred, parser=parser)
        true = html.fromstring(true, parser=parser)
        # print("pred:",pred)
        # print("true:",true)
        if pred.xpath('body/table') and true.xpath('body/table'):
            pred = pred.xpath('body/table')[0]
            true = true.xpath('body/table')[0]
            if self.ignore_nodes:
                etree.strip_tags(pred, *self.ignore_nodes)
                etree.strip_tags(true, *self.ignore_nodes)
            n_nodes_pred = len(pred.xpath('.//*'))
            n_nodes_true = len(true.xpath('.//*'))
            n_nodes = max(n_nodes_pred, n_nodes_true)
            tree_pred = self.load_html_tree(pred)
            tree_true = self.load_html_tree(true)
            distance = APTED(tree_pred, tree_true, CustomConfig()).compute_edit_distance()
            return 1.0 - (float(distance) / n_nodes)
        else:
            return 0.0

    def batch_evaluate(self, pred_json, true_json):
        """Computes TEDS score between the prediction and the ground truth of
        a batch of samples
        @params pred_json: {'FILENAME': 'HTML CODE', ...}
        @params true_json: {'FILENAME': {'html': 'HTML CODE'}, ...}
        @output: {'FILENAME': 'TEDS SCORE', ...}
        """
        samples = true_json.keys()
        if self.n_jobs == 1:
            scores = [
                self.evaluate(pred_json.get(filename, ''), true_json[filename]['html']) for filename in tqdm(samples)
            ]
        else:
            # inputs = [{'pred': pred_json.get(filename, ''), 'true': true_json[filename]['html']} for filename in samples]
            inputs = [{'pred': pred_json.get(filename, ''), 'true': true_json[filename]} for filename in samples]
            scores = parallel_process(inputs, self.evaluate, use_kwargs=True, n_jobs=self.n_jobs, front_num=1)
        scores = dict(zip(samples, scores))
        return scores


def convert_table_to_html_str(table_row_list=[]):
    """
    Given a list of table rows, build the corresponding html string, which is used to compute the TEDS score.
    We use the official code of PubTabNet to compute TEDS score, it does not consider '<th>' label.
    We also remove unneccessary spaces within a table cell and extra '\n' as they will influence the TEDS score.
    """
    html_table_str = '<html><body><table>' + '\n'
    for data_row in table_row_list:
        html_table_str += '<tr>'
        for cell_str in data_row:
            html_table_str += f'<td>{cell_str}</td>'
        html_table_str += '</tr>'
        html_table_str += '\n'
    html_table_str += '</table></body></html>'
    html_table_str = html_table_str.replace('\n', '')
    return html_table_str


def convert_markdown_table_to_html(markdown_table):
    """
    Converts a markdown table to the corresponding html string for TEDS computation.
    """
    # remove extra code block tokens like '```markdown' and '```
    markdown_table = markdown_table.strip('```markdown').strip('```').strip()
    row_str_list = markdown_table.split('\n')
    # extra the first header row and other data rows
    valid_row_str_list = [row_str_list[0]] + row_str_list[2:]
    table_rows = []
    for row_str in valid_row_str_list:
        one_row = []
        for cell in row_str.strip().split('|')[1:-1]:
            if set(cell) != set(' '):
                one_row.append(cell.strip())
            else:
                one_row.append(' ')
        table_rows.append(one_row)
    # build html string based on table rows
    html_str = convert_table_to_html_str(table_rows)
    return html_str


def dict_to_html(data):
    html = '<html><body><table>\n'
    for key, value in data.items():
        if not isinstance(value, str):
            value = str(value)
        value_str = ' '.join(value)

        html += f'  <tr><td>{key}</td><td>{value_str}</td></tr>\n'
    html += '</table></body></html>'
    return html


def convert_str_to_dict(predict_str: str):
    """
    Parses the 'predict' string and returns a dictionary.
    Missing or unparseable content is handled gracefully.

    Parameters:
    - predict_str (str): The prediction string containing the output dict.

    Returns:
    - dict: A dictionary extracted from the predict string.
    """
    # Remove code fences like ```python\n...\n```
    code_fence_pattern = r'```(?:python|json)?\n(.*?)\n```'
    match = re.search(code_fence_pattern, predict_str, re.DOTALL | re.IGNORECASE)
    if match:
        content = match.group(1)
    else:
        content = predict_str.strip()

    data = {}
    success = False

    # try parsing with JSON
    try:
        data = json.loads(content)
        success = True
    except json.JSONDecodeError:
        pass

    # try parsing with ast.literal_eval
    if not success:
        try:
            data = ast.literal_eval(content)
            if isinstance(data, dict):
                success = True
        except (ValueError, SyntaxError):
            pass

    # try parsing with regex
    if not success:
        key_value_pattern = r'["\']?([\w\s]+)["\']?\s*[:=]\s*["\']?([^\n,"\'{}]+)["\']?'
        matches = re.findall(key_value_pattern, content)
        try:
            for key, value in matches:
                data[key.strip()] = value.strip()
        except:
            return {}

    if not data:
        return {}

    try:
        result = {k.strip(): str(v).strip() for k, v in data.items()}
    except:
        return {}
    return result


def convert_str_to_multi_dict(predict_str: str):
    """
    Parses the 'predict' string and returns a dictionary.
    Handles nested dictionaries and missing or unparseable content gracefully.

    Parameters:
    - predict_str (str): The prediction string containing the output dict.

    Returns:
    - dict: A dictionary extracted from the predict string.
    """
    # Remove code fences like ```python\n...\n```
    code_fence_pattern = r'```(?:python|json)?\n(.*?)\n```'
    matches = re.findall(code_fence_pattern, predict_str, re.DOTALL | re.IGNORECASE)
    if matches:
        content = max(matches, key=len)
    else:
        content = predict_str.strip()

    def strip_variable_assignment(s):
        variable_assignment_pattern = r'^\s*\w+\s*=\s*'
        return re.sub(variable_assignment_pattern, '', s.strip(), count=1)

    content = strip_variable_assignment(content)

    def remove_comments(s):
        return re.sub(r'#.*', '', s)

    content = remove_comments(content)

    last_brace_pos = content.rfind('}')
    if last_brace_pos != -1:
        content = content[:last_brace_pos + 1]

    data = {}
    success = False

    # try parsing with ast.literal_eval
    try:
        data = ast.literal_eval(content)
        if isinstance(data, dict):
            success = True
    except (ValueError, SyntaxError, TypeError):
        pass

    if not success:
        return {}

    def process_data(obj):
        if isinstance(obj, dict):
            return {k: process_data(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [process_data(elem) for elem in obj]
        else:
            return obj

    data = process_data(data)

    return data


def generate_combinations(input_dict):
    """
    Function to generate all possible combinations of values from a dictionary.
    """
    kie_answer = input_dict
    if not isinstance(kie_answer, dict):
        kie_answer = kie_answer.strip('"')
        try:
            kie_answer = json.loads(kie_answer)
        except json.JSONDecodeError:
            try:
                kie_answer = ast.literal_eval(kie_answer)
                if not isinstance(kie_answer, dict):
                    kie_answer = ast.literal_eval(kie_answer)
            except (ValueError, SyntaxError):
                print(f"Unable to parse 'answers' field: {kie_answer}")
                return {}

        # Ensure the parsed result is a dictionary.
        if not isinstance(kie_answer, dict):
            print("Parsed 'answers' is still not a dictionary.")
            raise ValueError('Input could not be parsed into a dictionary.')

        keys = list(kie_answer.keys())

        value_lists = []
        for single_key in keys:
            sinlge_value = kie_answer[single_key]
            if not isinstance(sinlge_value, list):
                sinlge_value = [sinlge_value]
            value_lists.append(sinlge_value)

        # Compute the Cartesian product of the value lists.
        combinations = list(product(*value_lists))

        # Create a dictionary for each combination of values.
        result = [dict(zip(keys, values)) for values in combinations]

        return result

    else:
        keys = list(input_dict.keys())
        value_lists = [input_dict[key] for key in keys]

        # Compute the Cartesian product of the value lists.
        combinations = list(product(*value_lists))

        # Create a dictionary for each combination of values.
        result = [dict(zip(keys, values)) for values in combinations]

        return result


def compute_f1_score(preds, gts, ignores=[]):
    """Compute the F1-score for KIE task between predicted and ground truth dictionaries.

    Args:
        preds (dict): The predicted key-value pairs.
        gts (dict): The ground truth key-value pairs.
        ignores (list): The list of keys to ignore during evaluation.

    Returns:
        dict: A dictionary where keys are field names and values are their corresponding F1-scores.
    """
    # Optionally remove ignored keys from predictions and ground truths
    keys = set(preds.keys()).union(set(gts.keys())) - set(ignores)
    f1_scores = {}

    for key in keys:
        pred_value = preds.get(key, None)
        gt_value = gts.get(key, None)

        if pred_value:
            pred_value = pred_value.lower().strip().replace('\n', ' ').replace(' ', '')
        if gt_value:
            gt_value = gt_value.lower().strip().replace('\n', ' ').replace(' ', '')

        if pred_value is None and gt_value is None:
            continue
        elif pred_value is None:
            precision = 0.0
            recall = 0.0
        elif gt_value is None:
            # false positive
            precision = 0.0
            recall = 0.0
        else:
            if pred_value == gt_value:
                # True positive
                precision = 1.0
                recall = 1.0
            else:
                precision = 0.0
                recall = 0.0

        # Compute F1-score
        f1_score = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0
        f1_scores[key] = f1_score

    if len(f1_scores) == 0:
        return 0
    average_f1 = sum(f1_scores.values()) / len(f1_scores)

    return average_f1


def pre_clean(text):
    text = re.sub(r'<bos>|<eos>|<pad>|<unk>', '', text)
    text = re.sub(r'\s##(\S)', r'\1', text)
    text = re.sub(r'\\\s', r'\\', text)
    text = re.sub(r'\s\*\s\*\s', r'**', text)
    text = re.sub(r'{\s', r'{', text)
    text = re.sub(r'\s}', r'}', text)
    text = re.sub(r'\s}', r'}', text)
    text = re.sub(r'\\begin\s', r'\\begin', text)
    text = re.sub(r'\\end\s', r'\\end', text)
    text = re.sub(r'\\end{table}', r'\\end{table} \n\n', text)
    text = text.replace('\n', ' ')
    text = text.replace('*', ' ')
    text = text.replace('_', ' ')
    return text


def get_tree(input_str):
    tree = Node('ROOT').addkid(Node('TITLE'))

    lines = input_str.split('\n')
    lines = [pre_clean(line) for line in lines]
    last_title = ''
    for line in lines:
        if line.startswith('#'):
            child = tree.get('ROOT')
            line = line.replace('#', '')
            child.addkid(Node(line))
            last_title = line
        else:
            if last_title == '':
                child = tree.get('TITLE')
                child.addkid(Node(line))
            else:
                child = tree.get(last_title)
                child.addkid(Node(line))
    return tree


def STEDS(pred_tree, ref_tree):

    def my_distance(pred, ref):
        if len(pred.split()) == 0 or len(ref.split()) == 0:
            return 1
        else:
            return 0

    total_distance = simple_distance(pred_tree, ref_tree, label_dist=my_distance)
    num_of_nodes = max(len(list(pred_tree.iter())), len(list(ref_tree.iter())))
    return 1 - total_distance / num_of_nodes


def doc_parsing_evaluation(pred, gt):
    score = 0
    if not isinstance(pred, str):
        return 0
    pred_tree = get_tree(pred)
    gt_tree = get_tree(gt)
    score = STEDS(pred_tree, gt_tree)

    return score


def wrap_html_table(html_table):
    """
    The TEDS computation from PubTabNet code requires that the input html table should have <html>, <body>, and <table> tags.
    Add them if they are missing.
    """
    html_table = html_table.replace('\n', '')
    # add missing <table> tag if missing
    if '<table' in html_table and '</table>' not in html_table:
        html_table = html_table + '</table>'
    elif '<table' not in html_table and '</table>' in html_table:
        html_table = '<table>' + html_table
    elif '<table' not in html_table and '</table>' not in html_table:
        html_table = '<table>' + html_table + '</table>'
    else:
        pass
    # add <body> and <html> tags if missing
    if '<body>' not in html_table:
        html_table = '<body>' + html_table + '</body>'
    if '<html>' not in html_table:
        html_table = '<html>' + html_table + '</html>'
    return html_table


def get_anls(s1, s2):
    try:
        s1 = s1.lower()
        s2 = s2.lower()
    except:
        pass
    if s1 == s2:
        return 1.0
    iou = 1 - editdistance.eval(s1, s2) / max(len(s1), len(s2))
    anls = iou
    return anls


def ocr_eval(references, predictions):
    socre_ = 0.0
    None_num = 0
    for idx, ref_value in enumerate(references):
        pred_value = predictions[idx]
        pred_values, ref_values = [], []
        if isinstance(pred_value, str):
            pred_values.append(pred_value)
        else:
            pred_values = pred_value
        if isinstance(ref_value, str):
            ref_values.append(ref_value)
        else:
            ref_values = ref_value

        temp_score = 0.0
        temp_num = len(ref_values)

        for tmpidx, tmpref in enumerate(ref_values):
            tmppred = pred_values[tmpidx] if tmpidx < len(pred_values) else pred_values[0]
            if len(pred_values) == 1 and tmppred != 'None' and 'None' not in ref_values:  # pred 1, and not None
                temp_score = max(temp_score, get_anls(tmppred, tmpref))
                temp_num = len(ref_values)
            else:
                if tmppred == 'None' and tmpref != 'None':
                    temp_score += 0.0
                elif tmpref == 'None':
                    temp_num -= 1
                else:
                    temp_score += get_anls(tmppred, tmpref)
        if temp_num == 0:
            ocr_score = 0.0
            None_num += 1
        else:
            ocr_score = temp_score / (temp_num)
        socre_ += ocr_score
    if None_num == len(references):
        return 9999
    else:
        return round(socre_ / (len(references) - None_num), 5)


def csv_eval(predictions, references, easy, pred_type='json'):
    predictions = predictions
    labels = references

    def is_int(val):
        try:
            int(val)
            return True
        except ValueError:
            return False

    def is_float(val):
        try:
            float(val)
            return True
        except ValueError:
            return False

    def convert_dict_to_list(data):
        """
        Convert a dictionary to a list of tuples, handling both simple and nested dictionaries.

        Args:
        data (dict): The input dictionary, which might be nested or simple.

        Returns:
        list: A list of tuples generated from the input dictionary.
        """
        # print(data)
        converted_list = []
        for key, value in data.items():
            # Check if the value is a dictionary (indicating a nested structure)
            if isinstance(value, dict):
                # Handle nested dictionary
                for subkey, subvalue in value.items():
                    # converted_list.append((key, subkey, subvalue))
                    converted_list.append((key, subkey, re.sub(r'[^\d.-]', '', str(subvalue))))

            else:
                # Handle simple key-value pair
                # converted_list.append((key, "value", value))
                converted_list.append((key, 'value', re.sub(r'[^\d.-]', '', str(value))))
        return converted_list

    def csv2triples(csv, separator='\\t', delimiter='\\n'):
        lines = csv.strip().split(delimiter)
        header = lines[0].split(separator)
        triples = []
        for line in lines[1:]:
            if not line:
                continue
            values = line.split(separator)
            entity = values[0]
            for i in range(1, len(values)):
                if i >= len(header):
                    break
                # ---------------------------------------------------------
                temp = [entity.strip(), header[i].strip()]
                temp = [x if len(x) == 0 or x[-1] != ':' else x[:-1] for x in temp]
                value = values[i].strip()
                value = re.sub(r'[^\d.-]', '', str(value))
                # value = value.replace("%","")
                # value = value.replace("$","")
                triples.append((temp[0], temp[1], value))
                # ---------------------------------------------------------
        return triples

    def csv2triples_noheader(csv, separator='\\t', delimiter='\\n'):
        lines = csv.strip().split(delimiter)
        maybe_header = [x.strip() for x in lines[0].split(separator)]
        not_header = False
        if len(maybe_header) > 2:
            for c in maybe_header[1:]:
                try:
                    num = float(c)
                    not_header = True
                except:
                    continue
                if not_header:
                    break
        header = None if not_header else maybe_header
        data_start = 0 if not_header and separator in lines[0] else 1
        triples = []
        for line in lines[data_start:]:
            if not line:
                continue
            values = [x.strip() for x in line.split(separator)]
            entity = values[0]
            for i in range(1, len(values)):
                try:
                    temp = [entity if entity[-1] != ':' else entity[:-1], '']
                except:
                    temp = [entity, '']
                if header is not None:
                    try:
                        this_header = header[i]
                        temp = [entity, this_header]
                        temp = [x if x[-1] != ':' else x[:-1] for x in temp]
                    except:
                        this_header = entity.strip()
                value = values[i].strip()
                value = re.sub(r'[^\d.-]', '', str(value))
                # value = value.replace("%","")
                # value = value.replace("$","")
                triples.append((temp[0], temp[1], value))
                # ---------------------------------------------------------
        return triples

    def process_triplets(triplets):
        new_triplets = []
        for triplet in triplets:
            new_triplet = []
            triplet_temp = []
            if len(triplet) > 2:
                if is_int(triplet[2]) or is_float(triplet[2]):
                    triplet_temp = (triplet[0].lower(), triplet[1].lower(), float(triplet[2]))
                else:
                    triplet_temp = (triplet[0].lower(), triplet[1].lower(), triplet[2].lower())
            else:
                triplet_temp = (triplet[0].lower(), triplet[1].lower(), 'no meaning')
            new_triplets.append(triplet_temp)
        return new_triplets

    def intersection_with_tolerance(a, b, tol_word, tol_num):
        a = set(a)
        b = set(b)
        c = set()
        for elem1 in a:
            for elem2 in b:
                if is_float(elem1[-1]) and is_float(elem2[-1]):
                    if (((Levenshtein.distance(''.join(elem1[:-1]), ''.join(elem2[:-1])) <= tol_word) and
                         (abs(elem1[-1] - elem2[-1]) / (abs(elem2[-1]) + 0.000001) <= tol_num))
                        or ((''.join(elem1[:-1]) in ''.join(elem2[:-1])) and
                            (abs(elem1[-1] - elem2[-1]) / (abs(elem2[-1]) + 0.000001) <= tol_num))
                        or ((''.join(elem2[:-1]) in ''.join(elem1[:-1])) and
                            (abs(elem1[-1] - elem2[-1]) / (abs(elem2[-1]) + 0.000001) <= tol_num))):
                        c.add(elem1)
                else:
                    if Levenshtein.distance(
                        ''.join([str(i) for i in elem1]), ''.join([str(j) for j in elem2])
                    ) <= tol_word:
                        c.add(elem1)
        return list(c)

    def union_with_tolerance(a, b, tol_word, tol_num):
        c = set(a) | set(b)
        d = set(a) & set(b)
        e = intersection_with_tolerance(a, b, tol_word, tol_num)
        f = set(e)
        g = c - (f - d)
        return list(g)

    def get_eval_list(
        pred_csv, label_csv, separator='\\t', delimiter='\\n', tol_word=3, tol_num=0.05, pred_type='json'
    ):
        if pred_type == 'json':
            pred_triple_list = []
            for it in pred_csv:
                pred_triple_temp = convert_dict_to_list(it)
                pred_triple_pre = process_triplets(pred_triple_temp)
                pred_triple_list.append(pred_triple_pre)
        else:
            pred_triple_list = []
            for it in pred_csv:
                pred_triple_temp = csv2triples(it, separator=separator, delimiter=delimiter)
                # pred_triple_temp = csv2triples_noheader(it, separator=separator, delimiter=delimiter)
                pred_triple_pre = process_triplets(pred_triple_temp)
                pred_triple_list.append(pred_triple_pre)

        label_triple_list = []
        for it in label_csv:
            label_triple_temp = convert_dict_to_list(it)
            label_triple_pre = process_triplets(label_triple_temp)
            label_triple_list.append(label_triple_pre)

        intersection_list = []
        union_list = []
        sim_list = []
        # for each chart image
        for pred, label in zip(pred_triple_list, label_triple_list):
            for idx in range(len(pred)):
                try:
                    if label[idx][1] == 'value' and 'value' not in pred[idx][:2]:
                        pred[idx] = (pred[idx][0], 'value', pred[idx][2])
                    temp_pred_head = sorted(pred[idx][:2])
                    temp_gt_head = sorted(label[idx][:2])
                    pred[idx] = (temp_pred_head[0], temp_pred_head[1], pred[idx][2])
                    label[idx] = (temp_gt_head[0], temp_gt_head[1], label[idx][2])
                except:
                    continue
            intersection = intersection_with_tolerance(pred, label, tol_word=tol_word, tol_num=tol_num)
            union = union_with_tolerance(pred, label, tol_word=tol_word, tol_num=tol_num)
            sim = len(intersection) / len(union)
            intersection_list.append(intersection)
            union_list.append(union)
            sim_list.append(sim)
        return intersection_list, union_list, sim_list

    def get_ap(predictions, labels, sim_threhold, tolerance, separator='\\t', delimiter='\\n', easy=1):
        if tolerance == 'strict':
            tol_word = 0
            if easy == 1:
                tol_num = 0
            else:
                tol_num = 0.1

        elif tolerance == 'slight':
            tol_word = 2
            if easy == 1:
                tol_num = 0.05
            else:
                tol_num = 0.3

        elif tolerance == 'high':
            tol_word = 5
            if easy == 1:
                tol_num = 0.1
            else:
                tol_num = 0.5
        intersection_list, union_list, sim_list = get_eval_list(
            predictions,
            labels,
            separator=separator,
            delimiter=delimiter,
            tol_word=tol_word,
            tol_num=tol_num,
            pred_type=pred_type
        )
        ap = len([num for num in sim_list if num >= sim_threhold]) / (len(sim_list) + 1e-16)
        return ap

    map_strict = 0
    map_slight = 0
    map_high = 0
    s = '\\t'
    d = '\\n'

    for sim_threhold in np.arange(0.5, 1, 0.05):
        map_temp_strict = get_ap(
            predictions, labels, sim_threhold=sim_threhold, tolerance='strict', separator=s, delimiter=d, easy=easy
        )
        map_temp_slight = get_ap(
            predictions, labels, sim_threhold=sim_threhold, tolerance='slight', separator=s, delimiter=d, easy=easy
        )
        map_temp_high = get_ap(
            predictions, labels, sim_threhold=sim_threhold, tolerance='high', separator=s, delimiter=d, easy=easy
        )
        map_strict += map_temp_strict / 10
        map_slight += map_temp_slight / 10
        map_high += map_temp_high / 10

    em = get_ap(predictions, labels, sim_threhold=1, tolerance='strict', separator=s, delimiter=d, easy=easy)
    ap_50_strict = get_ap(
        predictions, labels, sim_threhold=0.5, tolerance='strict', separator=s, delimiter=d, easy=easy
    )
    ap_75_strict = get_ap(
        predictions, labels, sim_threhold=0.75, tolerance='strict', separator=s, delimiter=d, easy=easy
    )
    ap_90_strict = get_ap(
        predictions, labels, sim_threhold=0.90, tolerance='strict', separator=s, delimiter=d, easy=easy
    )
    ap_50_slight = get_ap(
        predictions, labels, sim_threhold=0.5, tolerance='slight', separator=s, delimiter=d, easy=easy
    )
    ap_75_slight = get_ap(
        predictions, labels, sim_threhold=0.75, tolerance='slight', separator=s, delimiter=d, easy=easy
    )
    ap_90_slight = get_ap(
        predictions, labels, sim_threhold=0.90, tolerance='slight', separator=s, delimiter=d, easy=easy
    )
    ap_50_high = get_ap(predictions, labels, sim_threhold=0.5, tolerance='high', separator=s, delimiter=d, easy=easy)
    ap_75_high = get_ap(predictions, labels, sim_threhold=0.75, tolerance='high', separator=s, delimiter=d, easy=easy)
    ap_90_high = get_ap(predictions, labels, sim_threhold=0.90, tolerance='high', separator=s, delimiter=d, easy=easy)

    return em, map_strict, map_slight, map_high, ap_50_strict, ap_75_strict, ap_90_strict, ap_50_slight, ap_75_slight, ap_90_slight, ap_50_high, ap_75_high, ap_90_high


def draw_SCRM_table(
    em,
    map_strict,
    map_slight,
    map_high,
    ap_50_strict,
    ap_75_strict,
    ap_90_strict,
    ap_50_slight,
    ap_75_slight,
    ap_90_slight,
    ap_50_high,
    ap_75_high,
    ap_90_high,
    title_ocr_socre,
    source_ocr_socre,
    x_title_ocr_socre,
    y_title_ocr_socre,
    structure_accuracy,
):
    result = f"""
            -----------------------------------------------------------\n
            |  Metrics   |  Sim_threshold  |  Tolerance  |    Value    |\n
            -----------------------------------------------------------\n
            |             |                 |   strict    |    {'%.4f' % map_strict}    |     \n
            |             |                 ----------------------------\n
            |  mPrecison  |  0.5:0.05:0.95  |   slight    |    {'%.4f' % map_slight}    |\n
            |             |                  ---------------------------\n
            |             |                 |    high     |    {'%.4f' % map_high}    |\n
            -----------------------------------------------------------\n
            |             |                 |   strict    |    {'%.4f' % ap_50_strict}    |\n
            |             |                  ---------------------------\n
            |  Precison   |       0.5       |   slight    |    {'%.4f' % ap_50_slight }    |\n
            |             |                  ---------------------------\n
            |             |                 |    high     |    {'%.4f' % ap_50_high }    |\n
            -----------------------------------------------------------\n
            |             |                 |   strict    |    {'%.4f' % ap_75_strict}    |\n
            |             |                  ---------------------------\n
            |  Precison   |      0.75       |   slight    |    {'%.4f' % ap_75_slight}    |\n
            |             |                  ---------------------------\n
            |             |                 |    high     |    {'%.4f' % ap_75_high}    |\n
            -----------------------------------------------------------\n
            |             |                 |   strict    |    {'%.4f' % ap_90_strict}    |\n
            |             |                  ---------------------------\n
            |  Precison   |       0.9       |   slight    |    {'%.4f' % ap_90_slight }    |\n
            |             |                  ---------------------------\n
            |             |                 |    high     |    {'%.4f' % ap_90_high}    |\n
            -----------------------------------------------------------\n
            |Precison(EM) |                                    {'%.4f' % em}    |\n
            -----------------------------------------------------------\n
            |Title(EM)    |                                    {'%.4f' % title_ocr_socre}    |\n
            -----------------------------------------------------------\n
            |Source(EM)   |                                    {'%.4f' % source_ocr_socre}    |\n
            -----------------------------------------------------------\n
            |X_title(EM)  |                                    {'%.4f' % x_title_ocr_socre}    |\n
            -----------------------------------------------------------\n
            |Y_title(EM)  |                                    {'%.4f' % y_title_ocr_socre}    |\n
            -----------------------------------------------------------\n
            |structure_acc|                                    {'%.4f' % structure_accuracy}    |\n
            -----------------------------------------------------------\n


            """
    return result


if __name__ == '__main__':
    import json
    import pprint

    # markdown structure for Table Parsing task
    pred_markdown = '| 1 | august 5 , 1972 | detroit lions | l 23 - 31 | 0 - 1 |\n| 2 | august 12 , 1972 | green bay packers | l 13 - 14 | 0 - 2 |\n| 3 | august 19 , 1972 | cincinnati bengals | w 35 - 17 | 1 - 2 |\n| 4 | august 25 , 1972 | atlanta falcons | w 24 - 10 | 2 - 2 |\n| 5 | august 31 , 1972 | washington redskins | l 24 - 27 | 2 - 3 |\n| 6 | september 10 , 1972 | minnesota vikings | w 21 - 19 | 3 - 3 |'
    true_markdown = '| week | date | opponent | result | record |\n| --- | --- | --- | --- | --- |\n| 1 | august 5 , 1972 | detroit lions | l 23 - 31 | 0 - 1 |\n| 2 | august 12 , 1972 | green bay packers | l 13 - 14 | 0 - 2 |\n| 3 | august 19 , 1972 | cincinnati bengals | w 35 - 17 | 1 - 2 |\n| 4 | august 25 , 1972 | atlanta falcons | w 24 - 10 | 2 - 2 |\n| 5 | august 31 , 1972 | washington redskins | l 24 - 27 | 2 - 3 |\n| 6 | september 10 , 1972 | minnesota vikings | w 21 - 19 | 3 - 3 |'
    teds = TEDS(n_jobs=4)
    pred_table_html = convert_markdown_table_to_html(pred_markdown)
    true_table_html = convert_markdown_table_to_html(true_markdown)

    scores = teds.evaluate(pred_table_html, true_table_html)

    pp = pprint.PrettyPrinter()
    pp.pprint(scores)

    # dict structure for Key Information Extraction task
    pred_dict = {'company': ['OLD TOWN '], 'date': ['2024'], 'address': ['SRI RAMPAI'], 'total': ['30']}
    true_dict = {
        'company': ['OLD TOWN KOPITAM SND BHD'],
        'date': ['2024/9/27'],
        'address': ['SRI RAMPAI'],
        'total': ['30']
    }
    teds = TEDS(n_jobs=4)
    pred_dict_html = dict_to_html(pred_dict)
    true_dict_html = dict_to_html(true_dict)
    print(pred_dict_html)
    print(true_dict_html)

    scores = teds.evaluate(pred_dict_html, true_dict_html)

    pp = pprint.PrettyPrinter()
    pp.pprint(scores)



import torch
from abc import abstractmethod
from typing import List

from ..model import ScoreModel


class VQAScoreModel(ScoreModel):

    @abstractmethod
    def forward(
        self, images: List[str], texts: List[str], question_template: str, answer_template: str
    ) -> torch.Tensor:
        """Forward pass of the model to return n scores for n (image, text) pairs (in PyTorch Tensor)
        question_template: a string with optional {} to be replaced with the 'text'
        answer_template: a string with optional {} to be replaced with the 'text'
        """
        pass



from typing import Any, Dict

from evalscope.api.benchmark import BenchmarkMeta, DefaultDataAdapter
from evalscope.api.dataset import Sample
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.logger import get_logger

logger = get_logger()


@register_benchmark(
    BenchmarkMeta(
        name='minerva_math',
        pretty_name='Minerva-Math',
        tags=[Tags.MATH, Tags.REASONING],
        description='Minerva-math is a benchmark designed to evaluate the mathematical and quantitative '
        'reasoning capabilities of LLMs. It consists of **272 problems** '
        'sourced primarily from **MIT OpenCourseWare** '
        'courses, covering advanced STEM subjects such as solid-state chemistry, astronomy, differential '
        'equations, and special relativity at the **university and graduate level**.',
        dataset_id='knoveleng/Minerva-Math',
        subset_list=['default'],
        metric_list=[{
            'acc': {
                'numeric': True
            }
        }],
        eval_split='train',
        prompt_template='{question}\nPlease reason step by step, and put your final answer within \\boxed{{}}.',
    )
)
class MinervaMathAdapter(DefaultDataAdapter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._use_llm_judge = True

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        return Sample(
            input=record['problem'],
            target=record['solution'],
            metadata={
                'type': record['type'],
                'idx': record['idx'],
            },
        )



from pydantic import BaseModel, Field, JsonValue
from typing import Dict, Literal, Optional, Sequence, Union


class ContentBase(BaseModel):
    internal: Optional[JsonValue] = Field(default=None)
    """Model provider specific payload - typically used to aid transformation back to model types."""


class ContentText(ContentBase):
    """Text content."""

    type: Literal['text'] = Field(default='text')
    """Type."""

    text: str
    """Text content."""

    refusal: Optional[bool] = Field(default=None)
    """Was this a refusal message?"""


class ContentReasoning(ContentBase):
    """Reasoning content.

    See the specification for [thinking blocks](https://docs.anthropic.com/en/docs/build-with-claude/extended-thinking#understanding-thinking-blocks) for Claude models.
    """  # noqa: E501

    type: Literal['reasoning'] = Field(default='reasoning')
    """Type."""

    reasoning: str
    """Reasoning content."""

    signature: Optional[str] = Field(default=None)
    """Signature for reasoning content (used by some models to ensure that reasoning content is not modified for replay)"""  # noqa: E501

    redacted: bool = Field(default=False)
    """Indicates that the explicit content of this reasoning block has been redacted."""


class ContentImage(ContentBase):
    """Image content."""

    type: Literal['image'] = Field(default='image')
    """Type."""

    image: str
    """Either a URL of the image or the base64 encoded image data."""

    detail: Literal['auto', 'low', 'high'] = Field(default='auto')
    """Specifies the detail level of the image.

    Currently only supported for OpenAI. Learn more in the    [Vision guide](https://platform.openai.com/docs/guides/vision/low-or-high-fidelity-image-understanding).
    """  # noqa: E501


class ContentAudio(ContentBase):
    """Audio content."""

    type: Literal['audio'] = Field(default='audio')
    """Type."""

    audio: str
    """Audio file path or base64 encoded data URL."""

    format: Literal['wav', 'mp3']
    """Format of audio data ('mp3' or 'wav')"""


class ContentVideo(ContentBase):
    """Video content."""

    type: Literal['video'] = Field(default='video')
    """Type."""

    video: str
    """Audio file path or base64 encoded data URL."""

    format: Literal['mp4', 'mpeg', 'mov']
    """Format of video data ('mp4', 'mpeg', or 'mov')"""


class ContentData(ContentBase):
    """Model internal."""

    type: Literal['data'] = Field(default='data')
    """Type."""

    data: Dict[str, JsonValue]
    """Model provider specific payload - required for internal content."""


Content = Union[
    ContentText,
    ContentReasoning,
    ContentImage,
    ContentAudio,
    ContentVideo,
    ContentData,
]
"""Content sent to or received from a model."""



import json
from typing import Any, Dict, List

from evalscope.api.benchmark import BenchmarkMeta, VisionLanguageAdapter
from evalscope.api.dataset import Sample
from evalscope.api.evaluator.state import TaskState
from evalscope.api.messages import ChatMessageUser, Content, ContentImage, ContentText
from evalscope.api.metric.scorer import Score
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.io_utils import bytes_to_base64
from evalscope.utils.logger import get_logger

logger = get_logger()

SUBSET_LIST = [
    'Regular Text Recognition', 'Irregular Text Recognition', 'Artistic Text Recognition', 'Handwriting Recognition',
    'Digit String Recognition', 'Non-Semantic Text Recognition', 'Scene Text-centric VQA', 'Doc-oriented VQA',
    'Key Information Extraction', 'Handwritten Mathematical Expression Recognition'
]


@register_benchmark(
    BenchmarkMeta(
        name='ocr_bench',
        pretty_name='OCRBench',
        tags=[Tags.MULTI_MODAL, Tags.KNOWLEDGE, Tags.QA],
        description=
        'OCRBench is a comprehensive evaluation benchmark designed to assess the OCR capabilities of Large Multimodal Models. It comprises five components: Text Recognition, SceneText-Centric VQA, Document-Oriented VQA, Key Information Extraction, and Handwritten Mathematical Expression Recognition. The benchmark includes 1000 question-answer pairs, and all the answers undergo manual verification and correction to ensure a more precise evaluation.',  # noqa: E501
        dataset_id='evalscope/OCRBench',
        subset_list=SUBSET_LIST,
        metric_list=['acc'],
        eval_split='test',
        prompt_template='{question}',
    )
)
class OCRBenchAdapter(VisionLanguageAdapter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_aggregation_name = False
        self.reformat_subset = True

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:

        input_text = self.prompt_template.format(question=record['question'])
        content_list: List[Content] = [ContentText(text=input_text)]
        image = record.get('image')
        if image:
            image_base64 = bytes_to_base64(image['bytes'], format='jpeg', add_header=True)
            content_list.append(ContentImage(image=image_base64))
        return Sample(
            input=[ChatMessageUser(content=content_list)],
            target=json.dumps(record.get('answer'), ensure_ascii=False),  # answers is a list
            subset_key=record.get('question_type'),
            metadata={
                'dataset': record.get('dataset'),
                'question_type': record.get('question_type'),
            }
        )

    def match_score(
        self, original_prediction: str, filtered_prediction: str, reference: str, task_state: TaskState
    ) -> Score:

        score = Score(
            extracted_prediction=filtered_prediction,
            prediction=original_prediction,
        )

        pred = filtered_prediction.lower().strip()
        gt_ans = json.loads(reference)
        dataset_name = task_state.metadata['dataset']

        score_value = 0
        if dataset_name == 'HME100k':
            if isinstance(gt_ans, list):
                for j in range(len(gt_ans)):
                    answer = gt_ans[j].strip().replace('\n', ' ').replace(' ', '')
                    predict = pred.strip().replace('\n', ' ').replace(' ', '')
                    if answer in predict:
                        score_value = 1
            else:
                answer = gt_ans.strip().replace('\n', ' ').replace(' ', '')
                predict = pred.strip().replace('\n', ' ').replace(' ', '')
                if answer in predict:
                    score_value = 1
        else:
            if isinstance(gt_ans, list):
                for j in range(len(gt_ans)):
                    answer = gt_ans[j].lower().strip().replace('\n', ' ')
                    predict = pred.lower().strip().replace('\n', ' ')
                    if answer in predict:
                        score_value = 1
            else:
                answer = gt_ans.lower().strip().replace('\n', ' ')
                predict = pred.lower().strip().replace('\n', ' ')
                if answer in predict:
                    score_value = 1
        score.value = {'acc': score_value}
        return score



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Union

from evalscope.api.registry import get_filter


class Filter(ABC):
    """
    Filter classes operate on a sample level.
    """

    def __init__(self, *args, **kwargs) -> None:
        """
        Can define custom behavior here, if an individual instantiation of a Filter class should have state.
        """

    @abstractmethod
    def apply(self, instance: List[str]) -> List[str]:

        return instance

    def __call__(self, instance: str) -> str:
        """
        Allows the filter to be called like a function.
        """
        return self.apply([instance])[0]


@dataclass
class FilterEnsemble:
    """
    FilterEnsemble creates a pipeline applying multiple filters.
    Its intended usage is to stack multiple post-processing steps in order.
    """

    name: str
    filters: List[Callable[[], Filter]]

    def apply(self, instance: List[str]) -> List[str]:

        for f in self.filters:
            # apply filters in sequence
            instance = f.apply(instance)

        return instance

    def __call__(self, instance: str) -> str:
        """
        Allows the filter ensemble to be called like a function.
        """
        return self.apply([instance])[0]


def build_filter_ensemble(name: str = 'default', filters: Dict[str, Any] = {}) -> FilterEnsemble:
    """
    Create a filtering pipeline.
    """
    filter_funcs = []
    for filter_name, filter_args in filters.items():
        filter_cls = get_filter(filter_name)
        if isinstance(filter_args, list):
            filter_function = filter_cls(*filter_args)
        elif isinstance(filter_args, dict):
            filter_function = filter_cls(**filter_args)
        else:
            # Assume single value for simple filters
            filter_function = filter_cls(filter_args)
        # add the filter as a pipeline step
        filter_funcs.append(filter_function)

    return FilterEnsemble(name=name, filters=filter_funcs)



import re
import string

_ARTICLES = re.compile(r'\b(a|an|the)\b', re.UNICODE)


def _answer_to_bags(answer):
    if isinstance(answer, (list, tuple)):
        raw_spans = answer
    else:
        raw_spans = [answer]
    normalized_spans = []
    token_bags = []
    for raw_span in raw_spans:
        normalized_span = _normalize(raw_span)
        normalized_spans.append(normalized_span)
        token_bags.append(set(normalized_span.split()))
    return normalized_spans, token_bags


def _is_number(text):
    try:
        float(text)
        return True
    except ValueError:
        return False


def _remove_articles(text):
    return _ARTICLES.sub(' ', text)


def _white_space_fix(text):
    return ' '.join(text.split())


def _remove_punc(text):
    exclude = set(string.punctuation)
    if not _is_number(text):
        return ''.join(ch for ch in text if ch not in exclude)
    else:
        return text


def _fix_number(text):
    return str(float(text)) if _is_number(text) else text


def _tokenize(text):
    return re.split(' |-', text)


def _normalize(answer):
    tokens = [
        _white_space_fix(_remove_articles(_fix_number(_remove_punc(token.lower())))) for token in _tokenize(answer)
    ]
    tokens = [token for token in tokens if token.strip()]
    normalized = ' '.join(tokens).strip()
    return normalized



import os
import torch
from typing import List

from ...constants import CACHE_DIR
from ..model import ScoreModel

CLIP_MODELS = [
    'openai:RN50', 'yfcc15m:RN50', 'cc12m:RN50', 'openai:RN101', 'yfcc15m:RN101', 'openai:RN50x4', 'openai:RN50x16',
    'openai:RN50x64', 'openai:ViT-B-32', 'laion400m_e31:ViT-B-32', 'laion400m_e32:ViT-B-32', 'laion2b_e16:ViT-B-32',
    'laion2b_s34b_b79k:ViT-B-32', 'datacomp_xl_s13b_b90k:ViT-B-32', 'datacomp_m_s128m_b4k:ViT-B-32',
    'commonpool_m_clip_s128m_b4k:ViT-B-32', 'commonpool_m_laion_s128m_b4k:ViT-B-32',
    'commonpool_m_image_s128m_b4k:ViT-B-32', 'commonpool_m_text_s128m_b4k:ViT-B-32',
    'commonpool_m_basic_s128m_b4k:ViT-B-32', 'commonpool_m_s128m_b4k:ViT-B-32', 'datacomp_s_s13m_b4k:ViT-B-32',
    'commonpool_s_clip_s13m_b4k:ViT-B-32', 'commonpool_s_laion_s13m_b4k:ViT-B-32',
    'commonpool_s_image_s13m_b4k:ViT-B-32', 'commonpool_s_text_s13m_b4k:ViT-B-32',
    'commonpool_s_basic_s13m_b4k:ViT-B-32', 'commonpool_s_s13m_b4k:ViT-B-32', 'metaclip_400m:ViT-B-32',
    'metaclip_fullcc:ViT-B-32', 'datacomp_s34b_b86k:ViT-B-32-256', 'openai:ViT-B-16', 'laion400m_e31:ViT-B-16',
    'laion400m_e32:ViT-B-16', 'laion2b_s34b_b88k:ViT-B-16', 'datacomp_xl_s13b_b90k:ViT-B-16',
    'datacomp_l_s1b_b8k:ViT-B-16', 'commonpool_l_clip_s1b_b8k:ViT-B-16', 'commonpool_l_laion_s1b_b8k:ViT-B-16',
    'commonpool_l_image_s1b_b8k:ViT-B-16', 'commonpool_l_text_s1b_b8k:ViT-B-16', 'commonpool_l_basic_s1b_b8k:ViT-B-16',
    'commonpool_l_s1b_b8k:ViT-B-16', 'dfn2b:ViT-B-16', 'metaclip_400m:ViT-B-16', 'metaclip_fullcc:ViT-B-16',
    'laion400m_e31:ViT-B-16-plus-240', 'laion400m_e32:ViT-B-16-plus-240', 'openai:ViT-L-14', 'laion400m_e31:ViT-L-14',
    'laion400m_e32:ViT-L-14', 'laion2b_s32b_b82k:ViT-L-14', 'datacomp_xl_s13b_b90k:ViT-L-14',
    'commonpool_xl_clip_s13b_b90k:ViT-L-14', 'commonpool_xl_laion_s13b_b90k:ViT-L-14',
    'commonpool_xl_s13b_b90k:ViT-L-14', 'metaclip_400m:ViT-L-14', 'metaclip_fullcc:ViT-L-14', 'dfn2b:ViT-L-14',
    'dfn2b_s39b:ViT-L-14', 'openai:ViT-L-14-336', 'laion2b_s32b_b79k:ViT-H-14', 'metaclip_fullcc:ViT-H-14',
    'metaclip_altogether:ViT-H-14', 'dfn5b:ViT-H-14', 'dfn5b:ViT-H-14-378', 'laion2b_s12b_b42k:ViT-g-14',
    'laion2b_s34b_b88k:ViT-g-14', 'laion2b_s39b_b160k:ViT-bigG-14', 'metaclip_fullcc:ViT-bigG-14',
    'laion2b_s12b_b32k:roberta-ViT-B-32', 'laion5b_s13b_b90k:xlm-roberta-base-ViT-B-32',
    'frozen_laion5b_s13b_b90k:xlm-roberta-large-ViT-H-14', 'laion400m_s13b_b51k:convnext_base',
    'laion2b_s13b_b82k:convnext_base_w', 'laion2b_s13b_b82k_augreg:convnext_base_w',
    'laion_aesthetic_s13b_b82k:convnext_base_w', 'laion_aesthetic_s13b_b82k:convnext_base_w_320',
    'laion_aesthetic_s13b_b82k_augreg:convnext_base_w_320', 'laion2b_s26b_b102k_augreg:convnext_large_d',
    'laion2b_s29b_b131k_ft:convnext_large_d_320', 'laion2b_s29b_b131k_ft_soup:convnext_large_d_320',
    'laion2b_s34b_b82k_augreg:convnext_xxlarge', 'laion2b_s34b_b82k_augreg_rewind:convnext_xxlarge',
    'laion2b_s34b_b82k_augreg_soup:convnext_xxlarge', 'laion2b_s13b_b90k:coca_ViT-B-32',
    'mscoco_finetuned_laion2b_s13b_b90k:coca_ViT-B-32', 'laion2b_s13b_b90k:coca_ViT-L-14',
    'mscoco_finetuned_laion2b_s13b_b90k:coca_ViT-L-14', 'laion400m_s11b_b41k:EVA01-g-14',
    'merged2b_s11b_b114k:EVA01-g-14-plus', 'merged2b_s8b_b131k:EVA02-B-16', 'merged2b_s4b_b131k:EVA02-L-14',
    'merged2b_s6b_b61k:EVA02-L-14-336', 'laion2b_s4b_b115k:EVA02-E-14', 'laion2b_s9b_b144k:EVA02-E-14-plus',
    'webli:ViT-B-16-SigLIP', 'webli:ViT-B-16-SigLIP-256', 'webli:ViT-B-16-SigLIP-i18n-256', 'webli:ViT-B-16-SigLIP-384',
    'webli:ViT-B-16-SigLIP-512', 'webli:ViT-L-16-SigLIP-256', 'webli:ViT-L-16-SigLIP-384', 'webli:ViT-SO400M-14-SigLIP',
    'webli:ViT-SO400M-16-SigLIP-i18n-256', 'webli:ViT-SO400M-14-SigLIP-378', 'webli:ViT-SO400M-14-SigLIP-384',
    'webli:ViT-B-32-SigLIP2-256', 'webli:ViT-B-16-SigLIP2', 'webli:ViT-B-16-SigLIP2-256', 'webli:ViT-B-16-SigLIP2-384',
    'webli:ViT-B-16-SigLIP2-512', 'webli:ViT-L-16-SigLIP2-256', 'webli:ViT-L-16-SigLIP2-384',
    'webli:ViT-L-16-SigLIP2-512', 'webli:ViT-SO400M-14-SigLIP2', 'webli:ViT-SO400M-14-SigLIP2-378',
    'webli:ViT-SO400M-16-SigLIP2-256', 'webli:ViT-SO400M-16-SigLIP2-384', 'webli:ViT-SO400M-16-SigLIP2-512',
    'webli:ViT-gopt-16-SigLIP2-256', 'webli:ViT-gopt-16-SigLIP2-384', 'datacomp1b:ViT-L-14-CLIPA',
    'datacomp1b:ViT-L-14-CLIPA-336', 'datacomp1b:ViT-H-14-CLIPA', 'laion2b:ViT-H-14-CLIPA-336',
    'datacomp1b:ViT-H-14-CLIPA-336', 'datacomp1b:ViT-bigG-14-CLIPA', 'datacomp1b:ViT-bigG-14-CLIPA-336',
    'v1:nllb-clip-base', 'v1:nllb-clip-large', 'v1:nllb-clip-base-siglip', 'mrl:nllb-clip-base-siglip',
    'v1:nllb-clip-large-siglip', 'mrl:nllb-clip-large-siglip', 'datacompdr:MobileCLIP-S1', 'datacompdr:MobileCLIP-S2',
    'datacompdr:MobileCLIP-B', 'datacompdr_lt:MobileCLIP-B', 'datacomp1b:ViTamin-S', 'datacomp1b:ViTamin-S-LTT',
    'datacomp1b:ViTamin-B', 'datacomp1b:ViTamin-B-LTT', 'datacomp1b:ViTamin-L', 'datacomp1b:ViTamin-L-256',
    'datacomp1b:ViTamin-L-336', 'datacomp1b:ViTamin-L-384', 'datacomp1b:ViTamin-L2', 'datacomp1b:ViTamin-L2-256',
    'datacomp1b:ViTamin-L2-336', 'datacomp1b:ViTamin-L2-384', 'datacomp1b:ViTamin-XL-256', 'datacomp1b:ViTamin-XL-336',
    'datacomp1b:ViTamin-XL-384', 'openai:RN50-quickgelu', 'yfcc15m:RN50-quickgelu', 'cc12m:RN50-quickgelu',
    'openai:RN101-quickgelu', 'yfcc15m:RN101-quickgelu', 'openai:RN50x4-quickgelu', 'openai:RN50x16-quickgelu',
    'openai:RN50x64-quickgelu', 'openai:ViT-B-32-quickgelu', 'laion400m_e31:ViT-B-32-quickgelu',
    'laion400m_e32:ViT-B-32-quickgelu', 'metaclip_400m:ViT-B-32-quickgelu', 'metaclip_fullcc:ViT-B-32-quickgelu',
    'openai:ViT-B-16-quickgelu', 'dfn2b:ViT-B-16-quickgelu', 'metaclip_400m:ViT-B-16-quickgelu',
    'metaclip_fullcc:ViT-B-16-quickgelu', 'openai:ViT-L-14-quickgelu', 'metaclip_400m:ViT-L-14-quickgelu',
    'metaclip_fullcc:ViT-L-14-quickgelu', 'dfn2b:ViT-L-14-quickgelu', 'openai:ViT-L-14-336-quickgelu',
    'metaclip_fullcc:ViT-H-14-quickgelu', 'dfn5b:ViT-H-14-quickgelu', 'dfn5b:ViT-H-14-378-quickgelu',
    'metaclip_fullcc:ViT-bigG-14-quickgelu'
]  # noqa: E501


class CLIPScoreModel(ScoreModel):
    "A wrapper for OpenCLIP models (including openAI's CLIP, OpenCLIP, DatacompCLIP)"

    def __init__(self, model_name='openai:ViT-L-14', device='cuda', cache_dir=CACHE_DIR):
        assert model_name in CLIP_MODELS
        super().__init__(model_name=model_name, device=device, cache_dir=cache_dir)

    def load_model(self):
        """Load the model, tokenizer, image transform
        """
        import open_clip

        from ..utils import download_open_clip_model

        self.pretrained, self.arch = self.model_name.split(':')
        # load model from modelscope
        model_file_path = download_open_clip_model(self.arch, self.pretrained, self.cache_dir)

        self.model, _, self.preprocess = open_clip.create_model_and_transforms(
            self.arch, pretrained=model_file_path, device=self.device
        )
        self.tokenizer = open_clip.get_tokenizer(self.arch)
        self.model.eval()

    def load_images(self, image: List[str]) -> torch.Tensor:
        """Load the image(s), and return a tensor (after preprocessing) put on self.device
        """
        image = [self.image_loader(x) for x in image]
        image = [self.preprocess(x) for x in image]
        image = torch.stack(image, dim=0).to(self.device)
        return image

    @torch.no_grad()
    def forward(self, images: List[str], texts: List[str]) -> torch.Tensor:
        """Forward pass of the model to return n scores for n (image, text) pairs (in PyTorch Tensor)
        """
        assert len(images) == len(texts)
        image = self.load_images(images)
        text = self.tokenizer(texts).to(self.device)
        image_features = self.model.encode_image(image)
        image_features /= image_features.norm(dim=-1, keepdim=True)
        text_features = self.model.encode_text(text)
        text_features /= text_features.norm(dim=-1, keepdim=True)

        # return cosine similarity as scores
        return (image_features * text_features).sum(dim=-1)



import inspect
from dataclasses import dataclass
from docstring_parser import Docstring, parse
from pydantic import BaseModel, Field, field_validator
from typing import Any, Callable, Dict, List, Literal, Optional, TypeAlias, Union, get_args, get_type_hints

from evalscope.utils.json_schema import JSONSchema, JSONType, json_schema, python_type_to_json_type

ToolParam: TypeAlias = JSONSchema
"""Description of tool parameter in JSON Schema format."""


class Tool:

    def __call__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        ...


class ToolParams(BaseModel):
    """Description of tool parameters object in JSON Schema format."""

    type: Literal['object'] = Field(default='object')
    """Params type (always 'object')"""

    properties: Dict[str, ToolParam] = Field(default_factory=dict)
    """Tool function parameters."""

    required: List[str] = Field(default_factory=list)
    """List of required fields."""

    additionalProperties: bool = Field(default=False)
    """Are additional object properties allowed? (always `False`)"""


@dataclass
class ToolDescription:
    name: Optional[str] = None
    description: Optional[str] = None
    parameters: Optional[ToolParams] = None


def tool_description(tool: Tool) -> ToolDescription:
    return getattr(tool, TOOL_DESCRIPTION, ToolDescription())


def set_tool_description(tool: Tool, description: ToolDescription) -> None:
    setattr(tool, TOOL_DESCRIPTION, description)


TOOL_DESCRIPTION = '__TOOL_DESCRIPTION__'


class ToolInfo(BaseModel):
    """Specification of a tool (JSON Schema compatible)

    If you are implementing a ModelAPI, most LLM libraries can
    be passed this object (dumped to a dict) directly as a function
    specification. For example, in the OpenAI provider:

    ```python
    ChatCompletionToolParam(
        type="function",
        function=tool.model_dump(exclude_none=True),
    )
    ```

    In some cases the field names don't match up exactly. In that case
    call `model_dump()` on the `parameters` field. For example, in the
    Anthropic provider:

    ```python
    ToolParam(
        name=tool.name,
        description=tool.description,
        input_schema=tool.parameters.model_dump(exclude_none=True),
    )
    ```
    """

    name: str
    """Name of tool."""
    description: str
    """Short description of tool."""
    parameters: ToolParams = Field(default_factory=ToolParams)
    """JSON Schema of tool parameters object."""
    options: Optional[Dict[str, object]] = Field(default=None)
    """Optional property bag that can be used by the model provider to customize the implementation of the tool"""


def parse_tool_info(func: Callable[..., Any]) -> ToolInfo:
    # tool may already have registry attributes w/ tool info
    description = tool_description(func)
    if (description.name and description.description and description.parameters is not None):
        return ToolInfo(
            name=description.name,
            description=description.description,
            parameters=description.parameters,
        )

    signature = inspect.signature(func)
    type_hints = get_type_hints(func)
    docstring = inspect.getdoc(func)
    parsed_docstring: Optional[Docstring] = parse(docstring) if docstring else None

    info = ToolInfo(name=func.__name__, description='')

    for param_name, param in signature.parameters.items():
        tool_param = ToolParam()

        # Parse docstring
        docstring_info = parse_docstring(docstring, param_name)

        # Get type information from type annotations
        if param_name in type_hints:
            tool_param = json_schema(type_hints[param_name])
        # as a fallback try to parse it from the docstring
        # (this is minimally necessary for backwards compatiblity
        #  with tools gen1 type parsing, which only used docstrings)
        elif 'docstring_type' in docstring_info:
            json_type = python_type_to_json_type(docstring_info['docstring_type'])
            if json_type and (json_type in get_args(JSONType)):
                tool_param = ToolParam(type=json_type)

        # Get default value
        if param.default is param.empty:
            info.parameters.required.append(param_name)
        else:
            tool_param.default = param.default

        # Add description from docstring
        if 'description' in docstring_info:
            tool_param.description = docstring_info['description']

        # append the tool param
        info.parameters.properties[param_name] = tool_param

    # Add function description if available
    if parsed_docstring:
        if parsed_docstring.description:
            info.description = parsed_docstring.description.strip()
        elif parsed_docstring.long_description:
            info.description = parsed_docstring.long_description.strip()
        elif parsed_docstring.short_description:
            info.description = parsed_docstring.short_description.strip()

        # Add examples if available
        if parsed_docstring.examples:
            examples = '\n\n'.join([(example.description or '') for example in parsed_docstring.examples])
            info.description = f'{info.description}\n\nExamples\n\n{examples}'

    return info


def parse_docstring(docstring: Optional[str], param_name: str) -> Dict[str, str]:
    if not docstring:
        return {}

    parsed_docstring: Docstring = parse(docstring)

    for param in parsed_docstring.params:
        if param.arg_name == param_name:
            schema: Dict[str, str] = {'description': param.description or ''}

            if param.type_name:
                schema['docstring_type'] = param.type_name

            return schema

    return {}



# Copyright (c) Alibaba, Inc. and its affiliates.
import glob
import importlib
import os
import time

from evalscope.utils import get_logger

logger = get_logger()

# Using glob to find all files matching the pattern
pattern = os.path.join(os.path.dirname(__file__), '*', '**', '*_adapter.py')
files = glob.glob(pattern, recursive=True)

import_times = []

for file_path in files:
    if file_path.endswith('.py') and not os.path.basename(file_path).startswith('_'):
        # Convert file path to a module path
        relative_path = os.path.relpath(file_path, os.path.dirname(__file__))
        module_path = relative_path[:-3].replace(os.path.sep, '.')  # strip '.py' and convert to module path
        full_path = f'evalscope.benchmarks.{module_path}'

        start_time = time.perf_counter()
        importlib.import_module(full_path)
        end_time = time.perf_counter()

        import_times.append((full_path, end_time - start_time))

# Sort by import time in descending order
import_times.sort(key=lambda x: x[1], reverse=True)

# Log the sorted import times
for module, duration in import_times:
    logger.debug(f'Module {module} imported in {duration:.6f} seconds')



import json
from pydantic import BaseModel, Field, JsonValue, field_validator
from typing import Any, Callable, Dict, List, Literal, Optional, Union


class ToolFunction(BaseModel):
    """Indicate that a specific tool function should be called."""

    name: str
    """The name of the tool function to call."""

    arguments: Dict[str, Any]
    """The arguments of the tool function to call"""

    @field_validator('arguments', mode='before')
    @classmethod
    def parse_arguments(cls, v):
        if isinstance(v, str):
            try:
                v = json.loads(v)
            except Exception as e:
                raise ValueError(f'arguments field string is not valid JSON: {e}')
        if not isinstance(v, dict):
            raise ValueError('arguments must be a dict or a JSON string representing a dict')
        return v


class ToolCallContent(BaseModel):
    """Content to include in tool call view."""

    title: Optional[str] = Field(default=None)
    """Optional (plain text) title for tool call content."""

    format: Literal['text', 'markdown']
    """Format (text or markdown)."""

    content: str
    """Text or markdown content."""


class ToolCallView(BaseModel):
    """Custom view of a tool call.

    Both `context` and `call` are optional. If `call` is not specified
    then the view will default to a syntax highlighted Python function call.
    """

    context: Optional[ToolCallContent] = Field(default=None)
    """Context for the tool call (i.e. current tool state)."""

    call: Optional[ToolCallContent] = Field(default=None)
    """Custom representation of tool call."""


class ToolCall(BaseModel):
    id: str
    """Unique identifier for tool call."""

    function: ToolFunction
    """Function to call."""

    internal: Optional[JsonValue] = Field(default=None)
    """Model provider specific payload - typically used to aid transformation back to model types."""

    parse_error: Optional[str] = Field(default=None)
    """Error which occurred parsing tool call."""

    view: Optional[ToolCallContent] = Field(default=None)
    """Custom view of tool call input."""

    type: Optional[str] = Field(default=None)
    """Tool call type (deprecated)."""


class ToolCallError(BaseModel):
    """Error raised by a tool call."""

    type: Literal[
        'parsing',
        'timeout',
        'unicode_decode',
        'permission',
        'file_not_found',
        'is_a_directory',
        'limit',
        'approval',
        'unknown',
    ]
    """Error type."""

    message: str
    """Error message."""


ToolChoice = Union[Literal['auto', 'any', 'none'], ToolFunction]
"""Specify which tool to call.

"auto" means the model decides; "any" means use at least one tool,
"none" means never call a tool; ToolFunction instructs the model
to call a specific function.
"""



from typing import Any, Dict, List

from evalscope.api.benchmark import BenchmarkMeta, VisionLanguageAdapter
from evalscope.api.dataset import Sample
from evalscope.api.evaluator.state import TaskState
from evalscope.api.messages import ChatMessageUser, Content, ContentImage, ContentText
from evalscope.api.metric.scorer import AggScore, SampleScore, Score
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.io_utils import bytes_to_base64
from evalscope.utils.logger import get_logger

logger = get_logger()


@register_benchmark(
    BenchmarkMeta(
        name='pope',
        pretty_name='POPE',
        tags=[Tags.MULTI_MODAL, Tags.HALLUCINATION, Tags.YES_NO],
        description=
        'POPE (Polling-based Object Probing Evaluation) is a benchmark designed to evaluate object hallucination in large vision-language models (LVLMs). It tests models by having them answer simple yes/no questions about the presence of specific objects in an image. This method helps measure how accurately a model\'s responses align with the visual content, with a focus on identifying instances where models claim objects exist that are not actually present. The benchmark employs various sampling strategies, including random, popular, and adversarial sampling, to create a robust set of questions for assessment.',  # noqa: E501
        dataset_id='lmms-lab/POPE',
        metric_list=['accuracy', 'precision', 'recall', 'f1_score', 'yes_ratio'],
        subset_list=['popular', 'adversarial', 'random'],
        default_subset='Full',
        prompt_template='{question}\nPlease answer YES or NO without an explanation.',
    )
)
class PopeAdapter(VisionLanguageAdapter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.split_as_subset = True
        self.add_overall_metric = False

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:

        input_text = self.prompt_template.format(question=record['question'])
        content_list: List[Content] = [ContentText(text=input_text)]
        image = record.get('image')
        if image:
            image_base64 = bytes_to_base64(image['bytes'], format='png', add_header=True)
            content_list.append(ContentImage(image=image_base64))
        answer = record['answer'].upper()  # 'YES' or 'NO'
        return Sample(
            input=[ChatMessageUser(content=content_list)],
            target=answer,
            metadata={
                'id': record.get('id'),
                'answer': answer,
                'category': record.get('category'),
                'question_id': record.get('question_id'),
            }
        )

    def match_score(self, original_prediction, filtered_prediction, reference, task_state) -> Score:
        score = Score(
            extracted_prediction=filtered_prediction,
            prediction=original_prediction,
        )
        # Check if the reference answer is in the filtered prediction
        result = 1 if reference in filtered_prediction.strip().upper() else 0
        score.value = {'acc': result}
        return score

    def aggregate_scores(self, sample_scores: List[SampleScore]) -> List[AggScore]:
        """
        Custom aggregation to compute accuracy, precision, recall, f1_score, and yes_ratio.
        """

        def compute_metrics(scores: List[SampleScore]):
            tp = fp = tn = fn = 0
            yes_count = 0
            total_count = len(scores)

            for ss in scores:
                gt = ss.sample_metadata['answer'].strip().upper()
                # Get prediction based on score
                pred = gt if ss.score.main_value == 1 else ('NO' if gt == 'YES' else 'YES')
                if pred == 'YES':
                    yes_count += 1
                if pred == 'YES' and gt == 'YES':
                    tp += 1
                elif pred == 'YES' and gt == 'NO':
                    fp += 1
                elif pred == 'NO' and gt == 'NO':
                    tn += 1
                elif pred == 'NO' and gt == 'YES':
                    fn += 1

            accuracy = (tp + tn) / total_count if total_count > 0 else 0.0
            precision = tp / (tp + fp) if (tp + fp) > 0 else 0.0
            recall = tp / (tp + fn) if (tp + fn) > 0 else 0.0
            f1_score = (2 * precision * recall) / (precision + recall) if (precision + recall) > 0 else 0.0
            yes_ratio = yes_count / total_count if total_count > 0 else 0.0

            return {
                'accuracy': accuracy,
                'precision': precision,
                'recall': recall,
                'f1_score': f1_score,
                'yes_ratio': yes_ratio
            }

        overall_metrics = compute_metrics(sample_scores)
        agg_scores = []
        for metric_name, value in overall_metrics.items():
            agg_scores.append(AggScore(metric_name=metric_name, score=value, num=len(sample_scores), metadata={}))

        return agg_scores



# flake8: noqa
import numpy as np
import os
import sys
from collections import defaultdict
from typing import Dict, List

from evalscope.utils import get_logger

logger = get_logger()


class End2EndEvaluator():

    def __init__(
        self,
        prediction: List,
        reference: List,
        metrics: Dict,
        match_method: str = 'quick_match',
        filter_types: dict = None
    ):

        self.match_method = match_method
        self.references = reference
        self.predictions = prediction
        self.dafault_metircs_dict = metrics

        filtered_gt_samples = []
        if filter_types:
            for gt_sample in self.references:
                select_flag = True
                for k, v in filter_types.items():
                    if gt_sample['page_info']['page_attribute'][k] != v:
                        select_flag = False
                if select_flag:
                    filtered_gt_samples.append(gt_sample)
        else:
            filtered_gt_samples = self.references  #[{},{},{}]
        self.references = filtered_gt_samples

    def score(self) -> dict:
        samples = self.get_matched_elements(self.references, self.predictions)
        metrics = self.process_generated_metric_results(samples)
        return metrics

    def get_page_elements(self, selected_annos):
        saved_element_dict = defaultdict(list)
        related_truncated = []
        truncated_all = {}
        for relation in selected_annos['extra']['relation']:  # Handle truncated text issues
            if relation['relation_type'] == 'truncated':
                truncated_all[relation['source_anno_id']] = ''
                truncated_all[relation['target_anno_id']] = ''
                exist_flag = False
                for merge_list in related_truncated:
                    if relation['source_anno_id'] in merge_list or relation[
                        'target_anno_id'] in merge_list:  # Consider cases where three text blocks may need to be merged
                        merge_list.append(relation['source_anno_id'])
                        merge_list.append(relation['target_anno_id'])
                        exist_flag = True
                if not exist_flag:
                    related_truncated.append([relation['source_anno_id'], relation['target_anno_id']])

        for item in selected_annos['layout_dets']:
            if item['anno_id'] not in truncated_all.keys():
                saved_element_dict[item['category_type']].append(item)
            else:
                truncated_all[item['anno_id']] = item

        for merge_list in related_truncated:
            text_block_list = [truncated_all[key] for key in merge_list]
            sorted_block = sorted(text_block_list, key=lambda x: x['order'])
            text = ''
            for block in sorted_block:
                text += block['text']
            merged_block = {
                'category_type': sorted_block[0]['category_type'],  # Directly use information from the first block
                'order': sorted_block[0]['order'],
                'anno_id': sorted_block[0]['anno_id'],
                'text': text,
                'merge_list': sorted_block
            }
            saved_element_dict[sorted_block[0]['category_type']].append(merged_block)

        return saved_element_dict

    def get_page_elements_list(self, gt_page_elements, category_list):
        element_list = []
        for category_type in category_list:
            if gt_page_elements.get(category_type):
                element_list.extend(gt_page_elements[category_type])
        return element_list

    def get_sorted_text_list(self, selected_annos):
        # txt_type: text, latex, html
        text_list = []
        for item in selected_annos:
            if item.get('order'):
                order = item['order']
            else:
                order = 0
            # 【txt_type,selecte_annos]
            text_list.append((order, item))
        sorted_text_list = sorted(text_list, key=lambda x: x[0])
        return [_[1] for _ in sorted_text_list]

    def filtered_out_ignore(self, items, ignore_category_list):
        filted_items = []
        for item in items:
            if item['gt_category_type'] not in ignore_category_list:
                filted_items.append(item)
        return filted_items

    def get_order_paired(self, order_match_s, img_name):
        matched = [(item['gt_position'], item['pred_position'])
                   for item in order_match_s
                   if (item['gt_position'] != [''] and item['pred_position'] != '')]
        gt_idx_all = [item['gt_position'] for item in order_match_s if (item['gt_position'] != [''])]
        read_order_pred = [i[0] for i in sorted(matched, key=lambda x: x[1])]
        read_order_gt = sum(gt_idx_all, [])  # Convert to one-dimensional list
        read_order_gt = [x for x in read_order_gt if x]
        gt = sorted(read_order_gt)
        pred = sum(read_order_pred, [])
        pred = [x for x in pred if x]
        if len(pred) > 0 or len(gt) > 0:
            import Levenshtein
            edit = Levenshtein.distance(gt, pred) / max(len(pred), len(gt))
            return {'gt': gt, 'pred': pred, 'img_id': img_name, 'edit': edit}
        else:
            return {}  # If both GT and pred are empty for the page, return empty

    def formula_format(self, formula_matches, img_name):
        # formated_list = []
        for i, item in enumerate(formula_matches):
            item['img_id'] = img_name + '_' + str(i)
        return formula_matches

    def get_matched_elements(self, references: list, predictions: list) -> dict:
        from .metrics import recogition_end2end_base_dataset, recogition_end2end_table_dataset

        plain_text_match = []
        display_formula_match = []
        html_table_match = []
        latex_table_match = []
        order_match = []

        for i, sample in enumerate(references):
            img_name = os.path.basename(sample['page_info']['image_path'])
            pred_content = predictions[i]
            result = self.process_get_matched_elements(sample, pred_content, img_name)
            [
                plain_text_match_clean, formated_display_formula, latex_table_match_s, html_table_match_s,
                order_match_single
            ] = result

            if order_match_single:
                order_match.append(order_match_single)
            if plain_text_match_clean:
                plain_text_match.extend(plain_text_match_clean)
            if formated_display_formula:
                display_formula_match.extend(formated_display_formula)
            if latex_table_match_s:
                latex_table_match.extend(latex_table_match_s)
            if html_table_match_s:
                html_table_match.extend(html_table_match_s)

        if len(latex_table_match) > len(html_table_match):
            table_match = latex_table_match
            table_format = 'latex'
        else:
            table_match = html_table_match
            table_format = 'html'

        matched_samples_all = {
            'text_block': recogition_end2end_base_dataset(plain_text_match),
            'display_formula': recogition_end2end_base_dataset(display_formula_match),
            'table': recogition_end2end_table_dataset(table_match, table_format),
            'reading_order': recogition_end2end_base_dataset(order_match)
        }

        return matched_samples_all

    def process_get_matched_elements(self, sample, pred_content, img_name):
        from func_timeout import FunctionTimedOut, func_timeout

        from .utils import match_gt2pred_no_split, match_gt2pred_quick, match_gt2pred_simple, md_tex_filter

        if self.match_method == 'simple_match':  # add match choice
            match_gt2pred = match_gt2pred_simple
        elif self.match_method == 'quick_match':
            match_gt2pred = match_gt2pred_quick
        elif self.match_method == 'no_split':
            match_gt2pred = match_gt2pred_no_split
        else:
            match_gt2pred = match_gt2pred_quick

        pred_dataset = md_tex_filter(pred_content)
        gt_page_elements = self.get_page_elements(sample)

        text_all = self.get_page_elements_list(
            gt_page_elements, [
                'text_block', 'title', 'code_txt', 'code_txt_caption', 'reference', 'equation_caption',
                'figure_caption', 'figure_footnote', 'table_caption', 'table_footnote', 'code_algorithm',
                'code_algorithm_caption', 'header', 'footer', 'page_footnote', 'page_number'
            ]
        )

        display_formula_match_s = []
        plain_text_match_clean = []
        latex_table_match_s = []
        html_table_match_s = []
        order_match_single = []
        if text_all:
            gt_text_list = self.get_sorted_text_list(text_all)
            try:
                plain_text_match_s = func_timeout(
                    30, match_gt2pred, args=(gt_text_list, pred_dataset['text_all'], 'text', img_name)
                )
            except FunctionTimedOut as e:
                logger.warning(f'Time out for plain text match of {img_name}, match_gt2pred_simple will be used.')
                plain_text_match_s = match_gt2pred_simple(gt_text_list, pred_dataset['text_all'], 'text', img_name)
                logger.error(str(e))
                raise e

            if not plain_text_match_s:
                logger.warning(f'No text match of {img_name}. The plain text match will be empty.')
            else:
                plain_text_match_clean = self.filtered_out_ignore(
                    plain_text_match_s, [
                        'figure_caption', 'figure_footnote', 'table_caption', 'table_footnote', 'code_algorithm',
                        'code_algorithm_caption', 'header', 'footer', 'page_footnote', 'page_number', 'equation_caption'
                    ]
                )

        if gt_page_elements.get('equation_isolated'):
            gt_display_list = self.get_sorted_text_list(gt_page_elements['equation_isolated'])
            display_formula_match_s = match_gt2pred(
                gt_display_list, pred_dataset['equation_isolated'], 'formula', img_name
            )
            display_formula_match_s = [x for x in display_formula_match_s if x['gt_idx'] != ['']]
            if not display_formula_match_s:
                logger.warning(f'No display_formula_match of {img_name}. The display_formula_match will be empty.')

        if gt_page_elements.get('table'):
            gt_table_list = self.get_sorted_text_list(gt_page_elements['table'])
            if pred_dataset['latex_table']:
                latex_table_match_s = match_gt2pred_simple(
                    gt_table_list, pred_dataset['latex_table'], 'latex_table', img_name
                )
                latex_table_match_s = [x for x in latex_table_match_s if x['gt_idx'] != ['']]
            if pred_dataset['html_table']:
                html_table_match_s = match_gt2pred_simple(
                    gt_table_list, pred_dataset['html_table'], 'html_table', img_name
                )
                html_table_match_s = [x for x in html_table_match_s if x['gt_idx'] != ['']]
            else:
                html_table_match_s = match_gt2pred_simple(gt_table_list, [], 'html_table', img_name)
                html_table_match_s = [x for x in html_table_match_s if x['gt_idx'] != ['']]

        order_match_s = plain_text_match_clean
        if order_match_s:
            order_match_single = self.get_order_paired(order_match_s, img_name)

        return [
            plain_text_match_clean, display_formula_match_s, latex_table_match_s, html_table_match_s, order_match_single
        ]

    def process_generated_metric_results(self, samples, save_name: str = 'end2end_quick_match'):
        from .metrics import METRIC_REGISTRY, get_full_labels_results, get_page_split, show_result

        result_all = {}
        page_info = {}
        metircs_dict = self.dafault_metircs_dict
        pages = self.references  #gt_samples list

        for page in pages:
            img_path = os.path.basename(page['page_info']['image_path'])
            page_info[img_path] = page['page_info']['page_attribute']

        for element in metircs_dict.keys():

            result = {}
            group_info = metircs_dict[element].get('group', [])
            # samples = samples.get(element) ##
            cur_samples = samples[element]

            for metric in metircs_dict[element]['metric']:
                metric_val = METRIC_REGISTRY.get(metric)

                cur_samples, result_s = metric_val(cur_samples).evaluate(group_info, f'{save_name}_{element}')
                if result_s:
                    result.update(result_s)

            if result:
                logger.info(f'{element}')
                show_result(result)
            result_all[element] = {}

            group_result = get_full_labels_results(cur_samples)
            page_result = get_page_split(cur_samples, page_info)

            result_all[element] = {'all': result, 'group': group_result, 'page': page_result}

        save_dict = {}
        en_overall = []
        ch_overall = []
        for category_type, metric in [('text_block', 'Edit_dist'), ('display_formula', 'Edit_dist'),
                                      ('display_formula', 'CDM'), ('table', 'TEDS'), ('table', 'Edit_dist'),
                                      ('reading_order', 'Edit_dist')]:
            if metric == 'TEDS':
                if category_type in result_all and 'page' in result_all[category_type] and metric in result_all[
                    category_type]['page']:
                    save_dict[category_type + '_' + metric
                              + '_EN'] = result_all[category_type]['page'][metric]['language: english']
                    save_dict[category_type + '_' + metric
                              + '_CH'] = result_all[category_type]['page'][metric]['language: simplified_chinese']
                else:
                    save_dict[category_type + '_' + metric + '_EN'] = np.nan
                    save_dict[category_type + '_' + metric + '_CH'] = np.nan
            else:
                if category_type in result_all and 'page' in result_all[category_type] and metric in result_all[
                    category_type]['page']:
                    save_dict[category_type + '_' + metric
                              + '_EN'] = result_all[category_type]['page'][metric].get('language: english', np.nan)
                    save_dict[category_type + '_' + metric + '_CH'] = result_all[category_type]['page'][metric].get(
                        'language: simplified_chinese', np.nan
                    )
                else:
                    save_dict[category_type + '_' + metric + '_EN'] = np.nan
                    save_dict[category_type + '_' + metric + '_CH'] = np.nan

            if metric == 'Edit_dist':
                if category_type in result_all and 'page' in result_all[category_type] and metric in result_all[
                    category_type]['page']:
                    en_overall.append(result_all[category_type]['page'][metric].get('language: english', np.nan))
                    ch_overall.append(
                        result_all[category_type]['page'][metric].get('language: simplified_chinese', np.nan)
                    )
                else:
                    en_overall.append(np.nan)
                    ch_overall.append(np.nan)

        en_overall_filtered = [x for x in en_overall if not np.isnan(x)]
        ch_overall_filtered = [x for x in ch_overall if not np.isnan(x)]
        save_dict['overall_EN'] = sum(en_overall_filtered) / len(en_overall_filtered) if en_overall_filtered else np.nan
        save_dict['overall_CH'] = sum(ch_overall_filtered) / len(ch_overall_filtered) if ch_overall_filtered else np.nan

        return save_dict



import ast
from typing import Any

from evalscope.api.benchmark import BenchmarkMeta, MultiChoiceAdapter
from evalscope.api.dataset import Sample
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.multi_choices import MultipleChoiceTemplate


@register_benchmark(
    BenchmarkMeta(
        name='musr',
        pretty_name='MuSR',
        tags=[Tags.REASONING, Tags.MULTIPLE_CHOICE],
        description=
        'MuSR is a benchmark for evaluating AI models on multiple-choice questions related to murder mysteries, object placements, and team allocation.',  # noqa: E501
        dataset_id='AI-ModelScope/MuSR',
        metric_list=['acc'],
        subset_list=['murder_mysteries', 'object_placements', 'team_allocation'],
        few_shot_num=0,
        train_split=None,
        eval_split='test',
        prompt_template=MultipleChoiceTemplate.SINGLE_ANSWER_COT,
    )
)
class MuSRAdapter(MultiChoiceAdapter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.split_as_subset = True

    def record_to_sample(self, record) -> Sample:
        choices = ast.literal_eval(record['choices'])
        choice_letters = ['A', 'B', 'C', 'D', 'E', 'F']
        target_letter = choice_letters[record['answer_index']]

        return Sample(
            input=f"{record['narrative']}\n\n{record['question']}",
            choices=choices,
            target=target_letter,
        )



# Copyright (c) Alibaba, Inc. and its affiliates.
import os
from collections import defaultdict

from evalscope.api.benchmark import BenchmarkMeta, MultiChoiceAdapter
from evalscope.api.dataset import Sample
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.io_utils import csv_to_list, jsonl_to_list
from evalscope.utils.logger import get_logger
from evalscope.utils.multi_choices import MultipleChoiceTemplate

# flake8: noqa

logger = get_logger()


@register_benchmark(
    BenchmarkMeta(
        name='general_mcq',
        pretty_name='General-MCQ',
        description='A general multiple-choice question answering dataset for custom evaluation. '
        'For detailed instructions on how to use this benchmark, please refer to the [User Guide](https://evalscope.readthedocs.io/en/latest/advanced_guides/custom_dataset/llm.html#mcq).',
        tags=[Tags.MULTIPLE_CHOICE, Tags.CUSTOM],
        dataset_id='general_mcq',
        subset_list=['default'],
        metric_list=['acc'],
        few_shot_num=0,
        train_split='dev',
        eval_split='val',
        prompt_template=MultipleChoiceTemplate.CHINESE_SINGLE_ANSWER_TEMPLATE,
    )
)
class GeneralMCQAdapter(MultiChoiceAdapter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.choices = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J']

    def load_from_disk(self, **kwargs):
        return super().load_from_disk(use_local_loader=True)

    def record_to_sample(self, record) -> Sample:
        # Extract choices from the record (A, B, C, D, etc.)
        choices = []
        for choice_key in self.choices:
            if choice_key in record:
                choices.append(record[choice_key])
            else:
                break  # Stop when we reach a choice key that doesn't exist

        return Sample(
            input=record['question'],
            choices=choices,
            target=record['answer'],
            metadata={'id': record.get('id', 'unknown')},
        )



import copy
from typing import TYPE_CHECKING, Callable, Dict, Optional, Type, Union

if TYPE_CHECKING:
    from evalscope.api.benchmark import BenchmarkMeta, DataAdapter
    from evalscope.api.filter import Filter
    from evalscope.api.metric import Aggregator, Metric
    from evalscope.api.model.model import ModelAPI
    from evalscope.config import TaskConfig

# BEGIN: Registry for benchmarks
# Registry for benchmarks, allowing dynamic registration and retrieval of benchmark metadata and data adapters.
BENCHMARK_REGISTRY: Dict[str, 'BenchmarkMeta'] = {}


def register_benchmark(metadata: 'BenchmarkMeta'):
    """Register a benchmark with its metadata."""

    def register_wrapper(data_adapter: Type['DataAdapter']):
        if metadata.name in BENCHMARK_REGISTRY:
            raise ValueError(f'Benchmark {metadata.name} already registered')
        metadata.data_adapter = data_adapter
        BENCHMARK_REGISTRY[metadata.name] = metadata
        return data_adapter

    return register_wrapper


def get_benchmark(name: str, config: Optional['TaskConfig'] = None) -> 'DataAdapter':
    """
    Retrieve a registered benchmark by name.

    Args:
        name (str): The name of the benchmark.
        config (Optional['TaskConfig']): The task configuration.
        dataset_args (Optional[dict]): The dataset-specific arguments.

    """
    # copy to avoid modifying the original metadata
    metadata = copy.deepcopy(BENCHMARK_REGISTRY.get(name))
    if not metadata:
        raise ValueError(f'Benchmark {name} not found, available benchmarks: {list(sorted(BENCHMARK_REGISTRY.keys()))}')

    # Update metadata with dataset-specific configuration
    if config is not None:
        metadata._update(config.dataset_args.get(name, {}))
    # Return the data adapter initialized with the benchmark metadata
    data_adapter_cls = metadata.data_adapter
    return data_adapter_cls(benchmark_meta=metadata, task_config=config)


# END: Registry for benchmarks

# BEGIN: Registry for model APIs
# Registry for model APIs, allowing dynamic registration and retrieval of model API classes.
MODEL_APIS: Dict[str, Type['ModelAPI']] = {}


def register_model_api(name: str):
    """
    Decorator to register a model API class with a given name.

    :param name: The name of the model API.
    """

    def decorator(api_class: Type['ModelAPI']):
        if name in MODEL_APIS:
            raise ValueError(f"Model API '{name}' is already registered.")
        MODEL_APIS[name] = api_class
        return api_class

    return decorator


def get_model_api(name: str) -> Type['ModelAPI']:
    """
    Retrieve a registered model API class by name.

    :param name: The name of the model API.
    :return: The model API class.
    """
    if name not in MODEL_APIS:
        raise ValueError(f"Model API '{name}' is not registered. Available model APIs: {list(MODEL_APIS.keys())}")

    wrapped = MODEL_APIS[name]
    if not isinstance(wrapped, type):
        return wrapped()
    else:
        return wrapped


# END: Registry for model APIs

# BEGIN: Registry for metrics
METRIC_REGISTRY: Dict[str, Type['Metric']] = {}


def register_metric(name: str):

    def decorate(fn):
        if name in METRIC_REGISTRY:
            raise ValueError(f"Metric named '{name}' conflicts with existing registered metric!")

        METRIC_REGISTRY[name] = fn
        return fn

    return decorate


def get_metric(name: str) -> Type['Metric']:
    if name in METRIC_REGISTRY:
        return METRIC_REGISTRY[name]
    else:
        raise ValueError(
            f"Metric '{name}' not found in the registry. Available metrics: {list(METRIC_REGISTRY.keys())}"
        )


# END: Registry for metrics

# BEGIN: Registry for filters

FILTER_REGISTRY: Dict[str, Type['Filter']] = {}


def register_filter(name):

    def decorate(cls):
        if name in FILTER_REGISTRY:
            raise ValueError(f'Registering filter `{name}` that is already in Registry {FILTER_REGISTRY}')
        FILTER_REGISTRY[name] = cls
        return cls

    return decorate


def get_filter(filter_name: str) -> Type['Filter']:
    if filter_name not in FILTER_REGISTRY:
        raise KeyError(
            f"Filter '{filter_name}' not found in the registry. Available filters: {list(FILTER_REGISTRY.keys())}"
        )
    return FILTER_REGISTRY[filter_name]


# END: Registry for filters

# BEGIN: Registry for aggregation functions
AGGREGATION_REGISTRY: Dict[str, Type['Aggregator']] = {}


def register_aggregation(name: str):
    """
    Decorator to register an aggregation function with a given name.

    :param name: The name of the aggregation function.
    """

    def decorator(aggregation_fn: 'Aggregator'):
        if name in AGGREGATION_REGISTRY:
            raise ValueError(f"Aggregation function '{name}' is already registered.")
        AGGREGATION_REGISTRY[name] = aggregation_fn
        return aggregation_fn

    return decorator


def get_aggregation(name: str) -> Type['Aggregator']:
    """
    Retrieve a registered aggregation function by name.

    :param name: The name of the aggregation function.
    :return: The aggregation function.
    """
    if name not in AGGREGATION_REGISTRY:
        raise ValueError(
            f"Aggregation function '{name}' is not registered. "
            f'Available aggregations: {list(AGGREGATION_REGISTRY.keys())}'
        )
    return AGGREGATION_REGISTRY[name]


# END: Registry for aggregation functions



import numpy as np
import os
import torch
from abc import ABC, abstractmethod
from PIL import Image
from typing import List

from ..constants import CACHE_DIR


def image_loader(image_path):
    if image_path.split('.')[-1] == 'npy':
        return Image.fromarray(np.load(image_path)[:, :, [2, 1, 0]], 'RGB')
    else:
        return Image.open(image_path).convert('RGB')


class ScoreModel(ABC):

    def __init__(self, model_name='clip-flant5-xxl', device='cuda', cache_dir=CACHE_DIR):
        self.model_name = model_name
        self.device = device
        self.cache_dir = cache_dir
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)
        self.image_loader = image_loader
        self.load_model()

    @abstractmethod
    def load_model(self):
        """Load the model, tokenizer, and etc.
        """
        pass

    @abstractmethod
    def load_images(self, image: List[str]) -> torch.Tensor:
        """Load the image(s), and return a tensor (after preprocessing) put on self.device
        """
        pass

    @abstractmethod
    def forward(self, images: List[str], texts: List[str], **kwargs) -> torch.Tensor:
        """Forward pass of the model to return n scores for n (image, text) pairs (in PyTorch Tensor)
        """
        pass



# flake8: noqa: E501
import glob
import os
from collections import defaultdict
from typing import Any, Dict, List

from evalscope.api.benchmark import BenchmarkMeta, DefaultDataAdapter
from evalscope.api.dataset import DatasetDict, DictDataLoader, Sample
from evalscope.api.evaluator import TaskState
from evalscope.api.messages.chat_message import ChatMessageUser
from evalscope.api.metric import AggScore, SampleScore, Score
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.report import Report, ReportKey
from evalscope.utils.logger import get_logger

logger = get_logger()

GRADER_SYSTEM_PROMPT = "Please act as an impartial judge and evaluate the quality of the responses provided by two AI assistants to the user prompt displayed below. You will be given assistant A's answer and assistant B's answer. Your job is to evaluate which assistant's answer is better.\n\nBegin your evaluation by generating your own answer to the prompt. You must provide your answers before judging any answers.\n\nWhen evaluating the assistants' answers, compare both assistants' answers with your answer. You must identify and correct any mistakes or inaccurate information.\n\nThen consider if the assistant's answers are helpful, relevant, and concise. Helpful means the answer correctly responds to the prompt or follows the instructions. Note when user prompt has any ambiguity or more than one interpretation, it is more helpful and appropriate to ask for clarifications or more information from the user than providing an answer based on assumptions. Relevant means all parts of the response closely connect or are appropriate to what is being asked. Concise means the response is clear and not verbose or excessive.\n\nThen consider the creativity and novelty of the assistant's answers when needed. Finally, identify any missing important information in the assistants' answers that would be beneficial to include when responding to the user prompt.\n\nAfter providing your explanation, you must output only one of the following choices as your final verdict with a label:\n\n1. Assistant A is significantly better: [[A>>B]]\n2. Assistant A is slightly better: [[A>B]]\n3. Tie, relatively the same: [[A=B]]\n4. Assistant B is slightly better: [[B>A]]\n5. Assistant B is significantly better: [[B>>A]]\n\nExample output: \"My final verdict is tie: [[A=B]]\"."  # noqa: E501

GRADER_TEMPLATE = "<|User Prompt|>\n{question}\n\n<|The Start of Assistant A's Answer|>\n{answer_1}\n<|The End of Assistant A's Answer|>\n\n<|The Start of Assistant B's Answer|>\n{answer_2}\n<|The End of Assistant B's Answer|>".strip(
)  # noqa: E501


@register_benchmark(
    BenchmarkMeta(
        name='general_arena',
        pretty_name='GeneralArena',
        tags=[Tags.CUSTOM, Tags.ARENA],
        description=
        'GeneralArena is a custom benchmark designed to evaluate the performance of large language models in a competitive setting, '
        'where models are pitted against each other in custom tasks to determine their relative strengths and weaknesses. You should '
        'provide the model outputs in the format of a list of dictionaries, where each dictionary contains the model name and its report path. '
        'For detailed instructions on how to use this benchmark, please refer to the [Arena User Guide](https://evalscope.readthedocs.io/en/latest/user_guides/arena.html).',
        dataset_id='general_arena',
        metric_list=['winrate'],
        few_shot_num=0,
        train_split=None,
        eval_split='test',
        system_prompt=GRADER_SYSTEM_PROMPT,
        prompt_template=GRADER_TEMPLATE,
        extra_params={
            'models': [{
                'name': 'qwen-plus',
                'report_path': 'outputs/20250627_172550/reports/qwen-plus'
            }, {
                'name': 'qwen2.5-7b',
                'report_path': 'outputs/20250627_172817/reports/qwen2.5-7b-instruct'
            }],
            'baseline':
            'qwen2.5-7b'
        }
    )
)
class GeneralArenaAdapter(DefaultDataAdapter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._use_llm_judge = True

        self.models = self.extra_params.get('models', [])
        self.baseline = self.extra_params.get('baseline', None)

    def load(self):
        """Load dataset by processing model reports."""
        self._check_names()
        self._check_reports()
        self._check_datasets()
        logger.info(f'Overall datasets: {self.overall_datasets}')
        dataset_model_dict = self._load_common_datasets()
        datasets = self._build_pair_wise_data(dataset_model_dict)

        # Convert to DatasetDict format
        dataset_dict = {}
        for subset_name, samples in datasets.items():
            dataset = DictDataLoader(
                dict_list=samples,
                limit=self.limit,
                shuffle=self.shuffle,
                repeats=self.repeats,
                sample_fields=self.record_to_sample
            ).load()
            dataset_dict[subset_name] = dataset

        test_dataset = DatasetDict(dataset_dict)
        return test_dataset, None

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        """Convert a data record to a Sample object."""
        return Sample(
            input=[ChatMessageUser(content=record['question'])],
            target=record['answer_2'],  # baseline answer
            metadata={
                'answer_1': record['answer_1'],
                'model_1': record['model_1'],
                'model_2': record['model_2'],
            }
        )

    def _check_names(self):
        """Check the names of the models and baseline."""
        # check duplicate models
        model_names = [model['name'] for model in self.models]
        if len(model_names) != len(set(model_names)):
            raise ValueError(f'Duplicate model names found in the models list {model_names}.')
        # check if models list is empty
        if len(self.models) < 2:
            raise ValueError('Models list must contain at least two models.')
        # check baseline model
        if self.baseline and self.baseline not in model_names:
            raise ValueError(f'Baseline model {self.baseline} not found in the models list.')
        # check if the baseline model is not set
        if not self.baseline:
            logger.warning('Baseline model is not set. Using the first model as the baseline.')
            self.baseline = self.models[0]['name']

    def _check_reports(self):
        """Check if the report paths are valid."""
        for model in self.models:
            report_path = model.get('report_path', None)
            if not report_path or not os.path.exists(report_path):
                raise ValueError(f'Report path {report_path} for model {model["name"]} does not exist.')
            reports = []
            for report_item in glob.glob(os.path.join(report_path, '*.json')):
                report = Report.from_json(report_item)
                reports.append(report)
            model['reports'] = reports

    def _check_datasets(self):
        """Check common datasets in the reports."""
        overall_datasets = set()
        for model in self.models:
            datasets = set()
            for report in model['reports']:
                report_df = report.to_dataframe()
                # get unique (dataset, subset) tuples
                unique_datasets = set(zip(report_df[ReportKey.dataset_name], report_df[ReportKey.subset_name]))
                datasets.update(unique_datasets)
            model['datasets'] = datasets
        # get overall datasets by intersecting all models' datasets
        overall_datasets = set.intersection(*[model['datasets'] for model in self.models if 'datasets' in model])
        self.overall_datasets = overall_datasets

    def _load_common_datasets(self):
        """Load common datasets from the local path."""
        from evalscope.utils import OutputsStructure
        from evalscope.utils.io_utils import jsonl_to_list

        dataset_dict = defaultdict(dict)
        for dataset_name, subset_name in self.overall_datasets:
            for model in self.models:
                dataset_path = model['report_path'].replace(OutputsStructure.REPORTS_DIR, OutputsStructure.REVIEWS_DIR)
                dataset_file_path = os.path.join(dataset_path, f'{dataset_name}_{subset_name}.jsonl')
                if not os.path.exists(dataset_file_path):
                    raise ValueError(
                        f'Dataset {dataset_name} with subset {subset_name} not found in model {model["name"]}.'
                    )
                dataset = jsonl_to_list(dataset_file_path)
                # sort by index
                dataset.sort(key=lambda x: x.get('index'))
                dataset_dict[(dataset_name, subset_name)][model['name']] = dataset

        return dataset_dict

    def _build_pair_wise_data(self, dataset_dict):
        """Build pairwise data for the models."""
        from evalscope.api.evaluator import ReviewResult
        from .utils import process_review_item

        pairwise_data = defaultdict(list)
        for (dataset_name, subset_name), model_data in dataset_dict.items():
            if len(model_data) < 2:
                logger.warning(f'Not enough models for dataset {dataset_name} with subset {subset_name}. Skipping.')
                continue
            # create pairwise data for each model against the baseline
            model_names = list(model_data.keys())
            for name in model_names:
                if name == self.baseline:
                    continue
                pairs = []
                for model_item, baseline_item in zip(model_data[name], model_data[self.baseline]):
                    # Convert to ReviewResult objects like in get_model_prediction
                    model_review = ReviewResult.model_validate(model_item)
                    baseline_review = ReviewResult.model_validate(baseline_item)

                    for model_choice, baseline_choice in zip(
                        process_review_item(model_review), process_review_item(baseline_review)
                    ):
                        pairs.append({
                            'question': model_choice['Question'],
                            'answer_1': model_choice['Generated'],
                            'answer_2': baseline_choice['Generated'],
                            'model_1': name,
                            'model_2': self.baseline
                        })
                pairwise_data[f'{dataset_name}&{subset_name}@{name}&{self.baseline}'] = pairs

        return pairwise_data

    def llm_match_score(
        self, original_prediction: str, filtered_prediction: str, reference: str, task_state: TaskState
    ) -> Score:
        """Use LLM as a judge to evaluate the predicted answer against the baseline."""
        from .utils import get_judge_score, post_process_result

        score = Score(
            extracted_prediction=filtered_prediction,
            prediction=original_prediction,
        )

        question = task_state.input_text
        answer_1 = task_state.metadata['answer_1']
        answer_2 = reference  # baseline answer
        model_1 = task_state.metadata['model_1']
        model_2 = task_state.metadata['model_2']

        system_template = self.system_prompt
        prompt_template = self.prompt_template

        prompt1 = prompt_template.format(question=question, answer_1=answer_1, answer_2=answer_2)
        # reverse the order
        prompt2 = prompt_template.format(question=question, answer_1=answer_2, answer_2=answer_1)

        # get grading response
        game1_response = self.llm_judge.judge(prompt1, system_prompt=system_template)
        game2_response = self.llm_judge.judge(prompt2, system_prompt=system_template)

        # parse grading response
        # game1
        res1 = post_process_result(game1_response)
        score1 = get_judge_score(res1, reverse=False)
        # game2
        res2 = post_process_result(game2_response)
        score2 = get_judge_score(res2, reverse=True)

        battle_result = {
            'score': (score1 + score2) / 2,
            'games': [
                {
                    'model_a': model_1,
                    'model_b': model_2,
                    'response': game1_response,
                    'judgment': res1
                },
                {
                    'model_a': model_2,
                    'model_b': model_1,
                    'response': game2_response,
                    'judgment': res2
                },
            ]
        }

        score.value = {'score': battle_result['score']}
        score.explanation = f'LLM judge battles: Game1: {game1_response[:100]}... Game2: {game2_response[:100]}...'
        score.metadata = {
            'source': 'llm_judge',
            'judge_strategy': getattr(self, 'judge_strategy', 'default'),
            'model': self.llm_judge.model_id if hasattr(self.llm_judge, 'model_id') else 'unknown',
            'battle_result': battle_result
        }
        score.main_score_name = 'score'

        return score

    def aggregate_scores(self, sample_scores: List[SampleScore]) -> List[AggScore]:
        """Aggregate scores to compute winrate."""
        import numpy as np
        import pandas as pd

        from .utils import compute_mle_elo, get_battles_from_row, get_bootstrap_result, get_win_rate_column

        battles = pd.concat([get_battles_from_row(res.score.metadata['battle_result']) for res in sample_scores])

        bt_model_coef = compute_mle_elo(battles, baseline_model=self.baseline)

        bootstrap_model_coef = get_bootstrap_result(
            battles, func_compute_elo=compute_mle_elo, num_round=100, baseline_model=self.baseline
        )

        stats = pd.DataFrame()
        stats['results'] = None
        stats['results'] = stats['results'].astype('object')

        for i, model in enumerate(bt_model_coef.index):
            stats.at[i, 'model'] = model
            stats.at[i, 'score'] = bt_model_coef[model]
            stats.at[i, 'lower'] = np.percentile(bootstrap_model_coef[model], 2.5)
            stats.at[i, 'upper'] = np.percentile(bootstrap_model_coef[model], 97.5)

        metrics_dict = {}
        metrics_dict['winrate'] = get_win_rate_column(stats, 'score', self.baseline).to_dict()
        metrics_dict['winrate_lower'] = get_win_rate_column(stats, 'lower', self.baseline).to_dict()
        metrics_dict['winrate_upper'] = get_win_rate_column(stats, 'upper', self.baseline).to_dict()

        agg_scores = []
        for metric_name, models in metrics_dict.items():
            for model_name, score_val in models.items():
                if model_name == self.baseline:
                    continue
                agg_scores.append(AggScore(score=score_val, metric_name=metric_name, num=len(sample_scores)))

        return agg_scores

    def extract_answer(self, prediction, task_state):
        # NOTE: This is a hacky way to extract the answer from the prediction
        return task_state.metadata['answer_1']

    def _on_generate_report_end(self, report: 'Report', output_dir: str, **kwargs):
        """Post-process the report to convert it to a DataFrame with winrate leaderboards."""
        import pandas as pd
        import tabulate

        report_path = output_dir
        leaderboard_file = os.path.join(report_path, 'leaderboard.txt')

        # Ensure report directory exists
        os.makedirs(report_path, exist_ok=True)

        # Convert report to dataframe
        df = report.to_dataframe()

        # Filter for winrate-related metrics
        winrate_df = df[df[ReportKey.metric_name].str.contains('winrate')].copy()

        if winrate_df.empty:
            logger.warning('No winrate data found in the report.')
            return

        # Get all model names from self.models
        all_model_names = [model['name'] for model in self.models]

        # Collect all leaderboard outputs
        leaderboard_outputs = []

        def format_leaderboard(data_df, title):
            """Format DataFrame as leaderboard with CI."""
            # Pivot to get winrate, winrate_lower, winrate_upper as columns
            pivot_df = data_df.pivot_table(
                index=[ReportKey.model_name], columns=ReportKey.metric_name, values=ReportKey.score, aggfunc='first'
            )

            # Add baseline model with 50% winrate
            baseline_data = {'winrate': 0.5, 'winrate_lower': 0.5, 'winrate_upper': 0.5}

            # Create a complete index with all models
            complete_index = pd.Index(all_model_names, name=pivot_df.index.name)
            pivot_df = pivot_df.reindex(complete_index)

            # Fill baseline model data
            if self.baseline in pivot_df.index:
                for col, val in baseline_data.items():
                    if col in pivot_df.columns:
                        pivot_df.loc[self.baseline, col] = val

            # Fill missing values with winrate score for other models
            if 'winrate' in pivot_df.columns:
                pivot_df['winrate_lower'] = pivot_df.get('winrate_lower', pivot_df['winrate'])
                pivot_df['winrate_upper'] = pivot_df.get('winrate_upper', pivot_df['winrate'])

            # Format for display
            leaderboard_data = []
            for model in pivot_df.index:
                if pd.isna(pivot_df.loc[model, 'winrate']):
                    continue

                score_pct = pivot_df.loc[model, 'winrate'] * 100
                lower_diff = (pivot_df.loc[model, 'winrate_lower'] - pivot_df.loc[model, 'winrate']) * 100
                upper_diff = (pivot_df.loc[model, 'winrate_upper'] - pivot_df.loc[model, 'winrate']) * 100

                leaderboard_data.append({
                    'Model': model,
                    'WinRate (%)': f'{score_pct:.1f}',
                    'CI (%)': f'({lower_diff:+.1f} / {upper_diff:+.1f})'
                })

            # Sort by score descending
            leaderboard_data.sort(key=lambda x: float(x['WinRate (%)'].replace('%', '')), reverse=True)

            # Create DataFrame
            leaderboard_df = pd.DataFrame(leaderboard_data)
            leaderboard_df.index = range(len(leaderboard_df))

            # Format as string
            table_str = tabulate.tabulate(leaderboard_df, headers='keys', showindex=False)
            output = f'{title}\n{table_str}\n'

            logger.info(f'\n{title}\n{table_str}')
            return output

        # Parse dataset and subset information from dataset_name column
        # Format: '{dataset_name}&{subset_name}@{name}&{self.baseline}'
        def parse_dataset_key(dataset_key):
            """Parse dataset key to extract dataset_name, subset_name, and model pair."""
            parts = dataset_key.split('@')

            dataset_subset = parts[0]
            model_pair = parts[1]

            dataset_name, subset_name = dataset_subset.split('&', 1)
            model_1, model_2 = model_pair.split('&', 1)

            return dataset_name, subset_name, model_1, model_2

        # Add parsed columns
        parsed_data = []
        for _, row in winrate_df.iterrows():
            dataset_name, subset_name, model_1, model_2 = parse_dataset_key(row[ReportKey.subset_name])
            if dataset_name is not None:
                parsed_data.append({
                    'dataset_name': dataset_name,
                    'subset_name': subset_name,
                    ReportKey.model_name: model_1,
                    ReportKey.metric_name: row[ReportKey.metric_name],
                    ReportKey.score: row[ReportKey.score]
                })

        if not parsed_data:
            logger.warning('No valid dataset keys found for parsing.')
            return

        parsed_df = pd.DataFrame(parsed_data)

        # 1. Overall ranking (aggregate across all datasets and subsets)
        overall_df = parsed_df.groupby([ReportKey.model_name,
                                        ReportKey.metric_name])[ReportKey.score].mean().reset_index()
        leaderboard_outputs.append(format_leaderboard(overall_df, '=== OVERALL LEADERBOARD ==='))

        # 2. Dataset-level rankings
        datasets = parsed_df['dataset_name'].unique()
        for dataset in sorted(datasets):
            dataset_df = parsed_df[parsed_df['dataset_name'] == dataset]
            dataset_agg = dataset_df.groupby([ReportKey.model_name,
                                              ReportKey.metric_name])[ReportKey.score].mean().reset_index()
            leaderboard_outputs.append(format_leaderboard(dataset_agg, f'=== DATASET LEADERBOARD: {dataset} ==='))

        # 3. Subset-level rankings
        subsets = parsed_df[['dataset_name', 'subset_name']].drop_duplicates()
        for _, subset_row in subsets.iterrows():
            dataset_name = subset_row['dataset_name']
            subset_name = subset_row['subset_name']
            subset_df = parsed_df[(parsed_df['dataset_name'] == dataset_name)
                                  & (parsed_df['subset_name'] == subset_name)]
            leaderboard_outputs.append(
                format_leaderboard(subset_df, f'=== SUBSET LEADERBOARD: {dataset_name} - {subset_name} ===')
            )

        # Write all leaderboard outputs to file
        with open(leaderboard_file, 'w', encoding='utf-8') as f:
            f.write('\n'.join(leaderboard_outputs))

        logger.info(f'Leaderboard results saved to: {leaderboard_file}')



# Copyright (c) Alibaba, Inc. and its affiliates.

import os
from typing import Any, Dict

from evalscope.api.benchmark import BenchmarkMeta, MultiChoiceAdapter
from evalscope.api.dataset import Sample
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.logger import get_logger
from evalscope.utils.multi_choices import FEW_SHOT_TEMPLATE, MultipleChoiceTemplate

logger = get_logger()

SUBSET_MAPPING = {
    'Electronic Science and Technology': ['Engineering'],
    'Philosophy': ['Philosophy'],
    'Traditional Chinese Medicine': ['Medicine'],
    'Applied Economics': ['Economics'],
    'Mathematics': ['Science'],
    'Physics': ['Science'],
    'Clinical Medicine': ['Medicine'],
    'Computer Science and Technology': ['Engineering'],
    'Information and Communication Engineering': ['Engineering'],
    'Control Science and Engineering': ['Engineering'],
    'Theoretical Economics': ['Economics'],
    'Law': ['Law'],
    'History': ['History'],
    'Basic Medicine': ['Medicine'],
    'Education': ['Education'],
    'Materials Science and Engineering': ['Engineering'],
    'Electrical Engineering': ['Engineering'],
    'Systems Science': ['Science'],
    'Power Engineering and Engineering Thermophysics': ['Engineering'],
    'Military Science': ['Military Science'],
    'Biology': ['Science'],
    'Business Administration': ['Management'],
    'Language and Literature': ['Literature and Arts'],
    'Public Health and Preventive Medicine': ['Medicine'],
    'Political Science': ['Law'],
    'Chemistry': ['Science'],
    'Hydraulic Engineering': ['Engineering'],
    'Chemical Engineering and Technology': ['Engineering'],
    'Pharmacy': ['Medicine'],
    'Geography': ['Science'],
    'Art Studies': ['Literature and Arts'],
    'Architecture': ['Engineering'],
    'Forestry Engineering': ['Engineering'],
    'Public Administration': ['Management'],
    'Oceanography': ['Science'],
    'Journalism and Communication': ['Literature and Arts'],
    'Nuclear Science and Technology': ['Engineering'],
    'Weapon Science and Technology': ['Engineering'],
    'Naval Architecture and Ocean Engineering': ['Engineering'],
    'Environmental Science and Engineering': ['Engineering'],
    'Transportation Engineering': ['Engineering'],
    'Geology': ['Science'],
    'Physical Oceanography': ['Science'],
    'Musicology': ['Literature and Arts'],
    'Stomatology': ['Medicine'],
    'Aquaculture': ['Agronomy'],
    'Mechanical Engineering': ['Engineering'],
    'Aeronautical and Astronautical Science and Technology': ['Engineering'],
    'Civil Engineering': ['Engineering'],
    'Mechanics': ['Engineering'],
    'Petroleum and Natural Gas Engineering': ['Engineering'],
    'Sociology': ['Sociology'],
    'Food Science and Engineering': ['Engineering'],
    'Agricultural Engineering': ['Engineering'],
    'Surveying and Mapping Science and Technology': ['Engineering'],
    'Metallurgical Engineering': ['Engineering'],
    'Library, Information and Archival Management': ['Management'],
    'Mining Engineering': ['Engineering'],
    'Astronomy': ['Science'],
    'Geological Resources and Geological Engineering': ['Engineering'],
    'Atmospheric Science': ['Science'],
    'Optical Engineering': ['Engineering'],
    'Animal Husbandry': ['Agronomy'],
    'Geophysics': ['Science'],
    'Crop Science': ['Agronomy'],
    'Management Science and Engineering': ['Management'],
    'Psychology': ['Education'],
    'Forestry': ['Agronomy'],
    'Textile Science and Engineering': ['Engineering'],
    'Veterinary Medicine': ['Agronomy'],
    'Instrument Science and Technology': ['Engineering'],
    'Physical Education': ['Education']
}


@register_benchmark(
    BenchmarkMeta(
        name='super_gpqa',
        pretty_name='SuperGPQA',
        tags=[Tags.KNOWLEDGE, Tags.MULTIPLE_CHOICE],
        description=
        'SuperGPQA is a large-scale multiple-choice question answering dataset, designed to evaluate the generalization ability of models across different fields. It contains 100,000+ questions from 50+ fields, with each question having 10 options.',  # noqa: E501
        dataset_id='m-a-p/SuperGPQA',
        subset_list=list(SUBSET_MAPPING.keys()),
        metric_list=['acc'],
        few_shot_num=0,
        train_split=None,
        eval_split='train',  # only have train split
        prompt_template=MultipleChoiceTemplate.SINGLE_ANSWER_COT,
    )
)
class SuperGPQAAdapter(MultiChoiceAdapter):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        if self.few_shot_num > 0 and self.few_shot_num != 5:
            logger.warning(
                f'Only support few_shot_num 0 or 5 for SuperGPQA, but got {self.few_shot_num}. Use 5-shot by default.'
            )
            self.few_shot_num = 5

        self.reformat_subset = True
        self.category_map = SUBSET_MAPPING

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        return Sample(
            input=record['question'],
            choices=record['options'],
            target=record['answer_letter'],
            subset_key=record['field'],
            metadata={
                'field': record['field'],
                'discipline': record['discipline'],
                'uuid': record.get('uuid', ''),
                'explanation': record.get('answer', ''),
            },
        )

    def format_fewshot_template(self, fewshot, sample):
        from .prompt import FEW_SHOT_SAMPLES

        return FEW_SHOT_TEMPLATE.format(fewshot=FEW_SHOT_SAMPLES, ) + self.format_prompt_template(sample)

    def extract_answer(self, prediction: str, task_state) -> str:
        """
        Extract the answer from the prediction.
        """
        from .utils import extract_option_content, extract_option_labels

        choices = [choice.value for choice in task_state.choices]
        if self.few_shot_num == 0:
            predict = extract_option_labels(prediction, 'ABCDEFGHIJ')
            if predict is None:
                # Try to extract by content matching
                predict = extract_option_content(prediction, choices)
                predict = chr(choices.index(predict) + 65) if predict else None
        else:
            response = prediction.split('Question:')[0]
            predict = extract_option_labels(response, 'ABCDEFGHIJ')
            if predict is None:
                predict = extract_option_content(response, choices)
                predict = chr(choices.index(predict) + 65) if predict else None
            if predict is None:
                predict = extract_option_labels(prediction, 'ABCDEFGHIJ')
                if predict is None:
                    predict = extract_option_content(prediction, choices)
                    predict = chr(choices.index(predict) + 65) if predict else None

        return predict or ''



from typing import Any, Dict, List

from evalscope.api.benchmark import BenchmarkMeta, VisionLanguageAdapter
from evalscope.api.dataset import Sample
from evalscope.api.evaluator.state import TaskState
from evalscope.api.messages.chat_message import ChatMessageUser
from evalscope.api.messages.content import Content, ContentImage, ContentText
from evalscope.api.metric.scorer import Score
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.io_utils import bytes_to_base64
from evalscope.utils.logger import get_logger

logger = get_logger()

SUBSET_LIST = [
    'OE_MM_maths_en_COMP',
    'OE_MM_maths_zh_CEE',
    'OE_MM_maths_zh_COMP',
    'OE_MM_physics_en_COMP',
    'OE_MM_physics_zh_CEE',
    'OE_TO_maths_en_COMP',
    'OE_TO_maths_zh_CEE',
    'OE_TO_maths_zh_COMP',
    'OE_TO_physics_en_COMP',
    'OE_TO_physics_zh_CEE',
    'TP_MM_maths_en_COMP',
    'TP_MM_maths_zh_CEE',
    'TP_MM_maths_zh_COMP',
    'TP_MM_physics_en_COMP',
    'TP_TO_maths_en_COMP',
    'TP_TO_maths_zh_CEE',
    'TP_TO_maths_zh_COMP',
    'TP_TO_physics_en_COMP',
]


@register_benchmark(
    BenchmarkMeta(
        name='olympiad_bench',
        pretty_name='OlympiadBench',
        tags=[Tags.MATH, Tags.REASONING],
        description='OlympiadBench is an Olympiad-level bilingual multimodal '
        'scientific benchmark, featuring 8,476 problems from '
        'Olympiad-level mathematics and physics competitions, '
        'including the Chinese college entrance exam. '
        'In the subsets: `OE` stands for `Open-Ended`, '
        '`TP` stands for `Theorem Proving`, '
        '`MM` stands for `Multimodal`, '
        '`TO` stands for `Text-Only`, '
        '`CEE` stands for `Chinese Entrance Exam`, '
        '`COMP` stands for `Comprehensive`. '
        '**Note: The `TP` subsets can\'t be evaluated with auto-judge for now**.',
        dataset_id='AI-ModelScope/OlympiadBench',
        subset_list=SUBSET_LIST,
        metric_list=['acc'],
        eval_split='train',
        prompt_template='{question}\nPlease reason step by step, and put your final answer within \\boxed{{}}.',
    )
)
class OlympiadBenchAdapter(VisionLanguageAdapter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        """Generate prompt for a single item."""
        from .utils import OlympiadBenchPrompter

        question = record.get('question', '')
        language = record.get('language', 'English')
        subject = record.get('subject', 'Math')
        question_type = record.get('question_type', '')
        answer_type = record.get('answer_type', '')
        is_multiple_answer = record.get('is_multiple_answer', False)
        unit = record.get('unit', '')
        # Generate prompt
        prompt = OlympiadBenchPrompter().make_prompt(
            problem=question,
            language=language,
            subject=subject,
            question_type=question_type,
            answer_type=answer_type,
            is_multiple_answer=is_multiple_answer,
            unit=unit,
        )
        # Construct content list
        content_list: List[Content] = []
        # Add images if available
        for i in range(9):
            image = record.get(f'image_{i+1}')
            if image:
                image_base64 = bytes_to_base64(image['bytes'], format='jpg', add_header=True)
                content_list.append(ContentImage(image=image_base64))
                prompt = prompt.replace(f'<image_{i+1}>', f'[image_{i+1}]')  # replace html tag
        # Add text content
        content_list.insert(0, ContentText(text=prompt))

        final_answer = record.get('final_answer', [])
        return Sample(
            input=[ChatMessageUser(content=content_list)],
            target=','.join(final_answer) if final_answer else '',
            metadata={
                'id': record.get('id', ''),
                'subfield': record.get('subfield', ''),
                'context': record.get('context', ''),
                'solution': record.get('solution', []),
                'final_answer': record.get('final_answer', []),
                'is_multiple_answer': is_multiple_answer,
                'unit': unit,
                'answer_type': answer_type,
                'question_type': question_type,
                'language': language,
                'subject': subject,
                'error': record.get('error', None),
            },
        )

    def extract_answer(self, prediction: str, task_state: TaskState):
        import re

        if task_state.metadata['language'] == 'Chinese':
            matches = re.findall('所以最终答案是(.*)', prediction)
        else:
            matches = re.findall('So the final answer is (.*)', prediction)

        # If found matches, take the last one, otherwise return the whole text
        if matches:
            return matches[-1].strip()
        return prediction

    def match_score(self, original_prediction, filtered_prediction, reference, task_state) -> Score:
        from .utils import MathJudger

        judger = MathJudger()
        score = Score(
            extracted_prediction=filtered_prediction,
            prediction=original_prediction,
        )
        question = task_state.metadata
        model_answer = filtered_prediction
        # Get precision/error threshold from reference if available
        answer_type = question['answer_type']
        try:
            if 'Tuple' in answer_type:  # 目前可机评的数据中 没有 need_human_evaluate
                judge_result = judger.judge(model_answer, question['final_answer'][0])
            else:
                if question['error']:
                    if ',' in question['error']:
                        precisions = question['error'].split(',')
                        precisions = [float(p) if p else 1e-8 for p in precisions]
                        judge_result = judger.judge(model_answer, question['final_answer'][0], precisions)
                    else:
                        precision = float(question['error'])
                        judge_result = judger.judge(model_answer, question['final_answer'][0], precision)
                else:
                    judge_result = judger.judge(model_answer, question['final_answer'][0])
        except Exception as e:
            logger.warning(f'Error in judging answer: {e}')
            judge_result = False

        score.value = {'acc': float(judge_result)}
        return score



import abc
from pydantic_core import to_jsonable_python
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Literal, Optional, Sequence, Union

from evalscope.api.messages import ChatMessage, ChatMessageAssistant, ChatMessageSystem, ChatMessageUser
from evalscope.api.registry import get_model_api
from evalscope.api.tool import ToolChoice, ToolFunction, ToolInfo
from evalscope.utils import get_logger
from evalscope.utils.function_utils import thread_safe
from .generate_config import GenerateConfig
from .model_output import ModelOutput

if TYPE_CHECKING:
    from evalscope.config import TaskConfig

logger = get_logger()


class ModelAPI(abc.ABC):
    """Model API provider."""

    def __init__(
        self,
        model_name: str,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        config: GenerateConfig = GenerateConfig(),
        **kwargs
    ) -> None:
        """Create a model API provider.

        Args:
           model_name (str): Model name.
           base_url (str | None): Alternate base URL for model.
           api_key (str | None): API key for model.
           api_key_vars (list[str]): Environment variables that
              may contain keys for this provider (used for override)
           config (GenerateConfig): Model configuration.
        """
        self.model_name = model_name
        self.base_url = base_url
        self.api_key = api_key
        self.config = config

    @abc.abstractmethod
    def generate(
        self,
        input: List[ChatMessage],
        tools: List[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        """Generate output from the model.

        Args:
          input (str | list[ChatMessage]): Chat message
            input (if a `str` is passed it is converted
            to a `ChatUserMessage`).
          tools (list[ToolInfo]): Tools available for the
            model to call.
          tool_choice (ToolChoice): Directives to the model
            as to which tools to prefer.
          config (GenerateConfig): Model configuration.

        Returns:
           ModelOutput
        """
        ...

    def batch_generate(
        self,
        inputs: List[List[ChatMessage]],
        tools: List[List[ToolInfo]],
        tool_choices: List[ToolChoice],
        configs: List[GenerateConfig],
    ) -> Generator[ModelOutput, None, None]:
        """Default batch implementation using individual generate calls.

        ModelAPI implementations can override this for optimized batch processing.

        Args:
          inputs: List of preprocessed chat message inputs.
          tools: List of tools for each input.
          tool_choices: List of tool choices for each input.
          configs: List of configs for each input.

        Returns:
            Generator yielding ModelOutput for each input.
        """
        from concurrent.futures import ThreadPoolExecutor

        def single_generate(args):
            input_msgs, input_tools, tool_choice, config = args
            return self.generate(input_msgs, input_tools, tool_choice, config)

        with ThreadPoolExecutor(max_workers=self.config.batch_size) as executor:
            futures = []
            for input_msgs, input_tools, tool_choice, config in zip(inputs, tools, tool_choices, configs):
                future = executor.submit(single_generate, (input_msgs, input_tools, tool_choice, config))
                futures.append(future)

            for future in futures:
                yield future.result()

    def supports_batch(self) -> bool:
        """Whether this ModelAPI supports optimized batch processing."""
        return False

    def max_tokens(self) -> Optional[int]:
        """Default max_tokens."""
        return None

    def max_tokens_for_config(self, config: GenerateConfig) -> Optional[int]:
        """Default max_tokens for a given config.

        Args:
           config: Generation config.

        Returns:
           Default maximum tokens for specified configuration.
        """
        return None

    def tools_required(self) -> bool:
        """Any tool use in a message stream means that tools must be passed."""
        return False

    def tool_result_images(self) -> bool:
        """Tool results can contain images"""
        return False


class Model:
    """Model interface.

    Use `get_model()` to get an instance of a model.
    """

    api: ModelAPI
    """Model API."""

    config: GenerateConfig
    """Generation config."""

    def __init__(self, api: ModelAPI, config: GenerateConfig, model_args: Dict[str, Any] = {}) -> None:
        """Create a model.

        Args:
           api: Model API provider.
           config: Model configuration.
           model_args: Optional model args
        """
        self.api = api
        self.config = config
        self.model_args = model_args

    @property
    def name(self) -> str:
        """Model name or path to model."""
        return self.api.model_name

    @property
    def role(self) -> Optional[str]:
        """Model role."""
        return self._role

    @role.setter
    def role(self, role: str) -> None:
        self._role = role

    def __str__(self) -> str:
        return f'Model(name={self.model_id}, role={self.role})'

    def generate(
        self,
        input: Union[str, List[ChatMessage]],
        tools: Optional[Sequence[ToolInfo]] = None,
        tool_choice: Optional[ToolChoice] = None,
        config: Optional[GenerateConfig] = None,
    ) -> ModelOutput:
        """Generate output from the model.

        Args:
          input: Chat message input (if a `str` is passed it is converted
            to a `ChatMessageUser`).
          tools: Tools available for the model to call.
          tool_choice: Directives to the model as to which tools to prefer.
          config: Model configuration.

        Returns:
           ModelOutput
        """
        processed_input, processed_tools, processed_tool_choice, processed_config = self._preprocess_input(
            input, tools, tool_choice, config
        )

        # Call the model's generate method
        output = self.api.generate(
            input=processed_input,
            tools=processed_tools,
            tool_choice=processed_tool_choice,
            config=processed_config,
        )

        # return output
        return output

    def batch_generate(
        self,
        inputs: List[List[ChatMessage]],
        tools: List[List[ToolInfo]],
        tool_choices: List[ToolChoice],
        configs: List[GenerateConfig],
    ) -> Generator[ModelOutput, None, None]:
        """Generate output from the model for a batch of inputs.

        Args:
          inputs (List[List[ChatMessage]]): Batch of chat message inputs.
          tools (List[List[ToolInfo]]): Batch of tools for each input.
          tool_choices (List[ToolChoice]): Batch of tool choices for each input.
          configs (List[GenerateConfig]): Batch of configs for each input.
        """
        preprocessed_data = []

        for input_item, input_tools, input_tool_choice, input_config in zip(inputs, tools, tool_choices, configs):
            processed_input, processed_tools, processed_tool_choice, processed_config = self._preprocess_input(
                input=input_item, tools=input_tools, tool_choice=input_tool_choice, config=input_config
            )
            preprocessed_data.append((processed_input, processed_tools, processed_tool_choice, processed_config))

        # check if ModelAPI supports batch processing
        if self.api.supports_batch() and len(preprocessed_data) > 1:
            # use the batch_generate method of the ModelAPI
            inputs, tools, tool_choices, configs = zip(*preprocessed_data)
            batch_results = self.api.batch_generate(
                inputs=list(inputs), tools=list(tools), tool_choices=list(tool_choices), configs=list(configs)
            )
            for result in batch_results:
                yield result
        else:
            # fall back to processing each input individually
            for input_msgs, input_tools, tool_choice, config in preprocessed_data:
                result = self.api.generate(input_msgs, input_tools, tool_choice, config)
                yield result

    def _preprocess_input(
        self,
        input: Union[str, List[ChatMessage]],
        tools: Optional[Sequence[ToolInfo]] = None,
        tool_choice: Optional[ToolChoice] = None,
        config: Optional[GenerateConfig] = None,
    ) -> tuple[List[ChatMessage], List[ToolInfo], ToolChoice, GenerateConfig]:
        """pre process input for generate."""

        # merge passed config
        if config is not None:
            config = self.config.merge(config)
        else:
            config = self.config.model_copy(deep=True)

        # provide max_tokens from the model api if required
        if config.max_tokens is None:
            config.max_tokens = self.api.max_tokens_for_config(config)
            if config.max_tokens is None:
                config.max_tokens = self.api.max_tokens()

        # normalize input to chat
        if isinstance(input, str):
            input = [ChatMessageUser(content=input)]

        # handle tools and tool_choice
        tool_choice = tool_choice if tool_choice is not None else 'auto'
        tools_info = list(tools) if tools is not None else []

        if isinstance(tool_choice, ToolFunction):
            tools_info = [tool for tool in tools_info if tool.name == tool_choice.name]

        if tool_choice == 'none' or len(tools_info) == 0:
            if not self.api.tools_required():
                tools_info = []
            tool_choice = 'none'

        return input, tools_info, tool_choice, config


class ModelCache:
    _models: Dict[str, 'Model'] = {}

    @classmethod
    def get(cls, key: str) -> Optional['Model']:
        return cls._models.get(key, None)

    @classmethod
    def set(cls, key: str, model: 'Model') -> None:
        cls._models[key] = model


def get_model_with_task_config(task_config: 'TaskConfig') -> Model:
    """Get an instance of a model with the specified task configuration.

    Args:
        task_config (TaskConfig): Task configuration.

    Returns:
        Model: An instance of the model.
    """
    model = task_config.model
    eval_type = task_config.eval_type
    base_url = task_config.api_url
    api_key = task_config.api_key
    config = task_config.generation_config
    model_args = task_config.model_args or {}

    return get_model(
        model=model, eval_type=eval_type, base_url=base_url, api_key=api_key, config=config, model_args=model_args
    )


@thread_safe
def get_model(
    model: Union[str, Model, ModelAPI],
    eval_type: str,
    base_url: Optional[str] = None,
    api_key: Optional[str] = None,
    config: GenerateConfig = GenerateConfig(),
    model_args: dict = {},
    role: Optional[str] = None,
    memoize: bool = True,
) -> Model:
    """Get an instance of a model.

    Calls to get_model() are memoized (i.e. a call with the same arguments
    will return an existing instance of the model rather than creating a
    new one). You can disable this with `memoize=False`.

    Args:
        task_config (TaskConfig): Task configuration.
        memoize (bool): Whether to memoize the model instance.

    Returns:
        Model instance.

    """

    # start with seeing if a model was passed
    if isinstance(model, Model):
        return model

    if isinstance(model, ModelAPI):
        return Model(model, config, model_args)

    # see if we can return a memoized model instance
    # (exclude mockllm since custom_outputs is an infinite generator)
    model_cache_key: str = ''
    if eval_type.startswith('mock_llm'):
        memoize = False
    if memoize:
        model_cache_key = (
            model + str(role) + config.model_dump_json(exclude_none=True) + str(base_url) + str(api_key)
            + str(to_jsonable_python(model_args, fallback=lambda _: None))
        )
        cached = ModelCache.get(model_cache_key)
        if cached is not None:
            return cached

    logger.info(
        f'Creating model {model} with eval_type={eval_type} '
        f'base_url={base_url}, config={config.model_dump(exclude_none=True)}, model_args={model_args}'
    )

    # find a matching model type
    modelapi_type = get_model_api(eval_type)

    modelapi_instance = modelapi_type(
        model_name=model,
        base_url=base_url,
        api_key=api_key,
        config=config,
        **model_args,
    )
    m = Model(modelapi_instance, config, model_args)
    if role is not None:
        m.role = role
    if memoize:
        ModelCache.set(model_cache_key, m)
    return m



def relaxed_correctness(prediction: str, target: str, max_relative_change: float = 0.05) -> bool:
    """Calculates relaxed correctness.

    The correctness tolerates certain error ratio defined by max_relative_change.
    See https://arxiv.org/pdf/2203.10244.pdf, end of section 5.1:
    “Following Methani et al. (2020), we use a relaxed accuracy measure for the
    numeric answers to allow a minor inaccuracy that may result from the automatic
    data extraction process. We consider an answer to be correct if it is within
    5% of the gold answer. For non-numeric answers, we still need an exact match
    to consider an answer to be correct.”

    This funcion is taken from https://github.com/QwenLM/Qwen-VL/blob/34b4c0ee7b07726371b960911f249fe61b362ca3/eval_mm/evaluate_vqa.py#L113
    Args:
      target: List of target string.
      prediction: List of predicted string.
      max_relative_change: Maximum relative change.

    Returns:
      Whether the prediction was correct given the specified tolerance.
    """  # noqa: E501

    def _to_float(text: str):
        try:
            if text.endswith('%'):
                # Convert percentages to floats.
                return float(text.rstrip('%')) / 100.0
            else:
                return float(text)
        except ValueError:
            return None

    prediction_float = _to_float(prediction)
    target_float = _to_float(target)
    if prediction_float is not None and target_float:
        relative_change = abs(prediction_float - target_float) / abs(target_float)
        return relative_change <= max_relative_change
    else:
        return prediction.lower() == target.lower()



# flake8: noqa: E501
from typing import Any, Dict, List

from evalscope.api.benchmark import BenchmarkMeta, VisionLanguageAdapter
from evalscope.api.dataset import Sample
from evalscope.api.messages import ChatMessageUser, Content, ContentImage, ContentText
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.io_utils import bytes_to_base64, compress_image_to_limit
from evalscope.utils.logger import get_logger

logger = get_logger()

# 定义提示模板
PROMPT_TEMPLATE = """{question}
\n\n\nLet's think step by step and give the final answer in curly braces,
like this: {{final answer}}"
"""

SUBSET_LIST = ['default']


@register_benchmark(
    BenchmarkMeta(
        name='zerobench',
        pretty_name='ZeroBench',
        dataset_id='evalscope/zerobench',
        tags=[Tags.KNOWLEDGE, Tags.QA, Tags.MULTI_MODAL],
        description=
        'ZeroBench is a challenging visual reasoning benchmark for Large Multimodal Models (LMMs). It consists of a main set of 100 high-quality, manually curated questions covering numerous domains, reasoning types and image type. Questions in ZeroBench have been designed and calibrated to be beyond the capabilities of current frontier models. As such, none of the evaluated models achieves a non-zero pass@1 (with greedy decoding) or 5/5 reliability score.',
        subset_list=SUBSET_LIST,
        metric_list=['acc'],
        eval_split='zerobench',
        train_split='zerobench_subquestions',
        prompt_template=PROMPT_TEMPLATE,
    )
)
class ZeroBenchAdapter(VisionLanguageAdapter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._use_llm_judge = True

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        question = record['question_text']
        content_list: List[Content] = [ContentText(text=self.prompt_template.format(question=question))]
        image = record['question_images_decoded']
        if len(image) > 0:
            for img in image:
                # Ensure image is under OpenAI's 10MB data-URI limit by compressing if needed
                processed_bytes, fmt = compress_image_to_limit(img['bytes'], 10_000_000)
                image_base64 = bytes_to_base64(processed_bytes, format=fmt, add_header=True)
                content_list.append(ContentImage(image=image_base64))

        metadata = {
            'question_id': record['question_id'],
            'question_images': record['question_images'],
            'image_attribution': record['image_attribution']
        }

        return Sample(
            input=[ChatMessageUser(content=content_list)], target=record['question_answer'], metadata=metadata
        )
